"""
Bases classes

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

import functools
import inspect
import itertools
import json
import logging
import math
import warnings
from abc import ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import numba as nb
import numpy as np
from numba.extending import is_jitted, overload, register_jitable

from ..tools.cache import cached_method, cached_property
from ..tools.docstrings import fill_in_docstring
from ..tools.misc import Number, hybridmethod
from ..tools.numba import jit
from ..tools.typing import (
    CellVolume,
    FloatNumerical,
    NumberOrArray,
    OperatorFactory,
    OperatorType,
)

if TYPE_CHECKING:
    from ._mesh import GridMesh
    from .boundaries.axes import Boundaries, BoundariesData  # @UnusedImport


PI_4 = 4 * np.pi
PI_43 = 4 / 3 * np.pi
CoordsType = Literal["cartesian", "grid", "cells"]


class OperatorInfo(NamedTuple):
    """stores information about an operator"""

    factory: OperatorFactory
    rank_in: int
    rank_out: int
    name: str = ""  # attach a unique name to help caching


def _check_shape(shape: Union[int, Sequence[int]]) -> Tuple[int, ...]:
    """checks the consistency of shape tuples"""
    if hasattr(shape, "__iter__"):
        shape_list: Sequence[int] = shape  # type: ignore
    else:
        shape_list = [shape]

    if len(shape_list) == 0:
        raise ValueError("Require at least one dimension")

    # convert the shape to a tuple of integers
    result = []
    for dim in shape_list:
        if dim == int(dim) and dim >= 1:
            result.append(int(dim))
        else:
            raise ValueError(f"{repr(dim)} is not a valid number of support points")
    return tuple(result)


def discretize_interval(
    x_min: float, x_max: float, num: int
) -> Tuple[np.ndarray, float]:
    r""" construct a list of equidistantly placed intervals 

    The discretization is defined as

    .. math::
            x_i &= x_\mathrm{min} + \left(i + \frac12\right) \Delta x
            \quad \text{for} \quad i = 0, \ldots, N - 1
        \\
            \Delta x &= \frac{x_\mathrm{max} - x_\mathrm{min}}{N}

    where :math:`N` is the number of intervals given by `num`.

    Args:
        x_min (float): Minimal value of the axis
        x_max (float): Maximal value of the axis
        num (int): Number of intervals

    Returns:
        tuple: (midpoints, dx): the midpoints of the intervals and the used
        discretization `dx`.
    """
    dx = (x_max - x_min) / num
    return (np.arange(num) + 0.5) * dx + x_min, dx


class DomainError(ValueError):
    """exception indicating that point lies outside domain"""

    pass


class DimensionError(ValueError):
    """exception indicating that dimensions were inconsistent"""

    pass


class PeriodicityError(RuntimeError):
    """exception indicating that the grid periodicity is inconsistent"""

    pass


class GridBase(metaclass=ABCMeta):
    """Base class for all grids defining common methods and interfaces"""

    _subclasses: Dict[str, Type[GridBase]] = {}  # all classes inheriting from this
    _operators: Dict[str, OperatorInfo] = {}  # all operators defined for the grid

    # properties that are defined in subclasses
    dim: int
    """int: The spatial dimension in which the grid is embedded"""
    axes: List[str]
    """list: Names of all axes that are described by the grid"""
    axes_symmetric: List[str] = []
    """list: The names of the additional axes that the fields do not depend on,
    e.g. along which they are constant. """

    boundary_names: Dict[str, Tuple[int, bool]] = {}
    """dict: Names of boundaries to select them conveniently"""
    cell_volume_data: Sequence[FloatNumerical]
    """list: Information about the size of discretization cells"""
    coordinate_constraints: List[int] = []
    """list: axes that not described explicitly"""
    num_axes: int
    """int: Number of axes that are *not* assumed symmetrically"""

    # mandatory, immutable, private attributes
    _axes_bounds: Tuple[Tuple[float, float], ...]
    _axes_coords: Tuple[np.ndarray, ...]
    _discretization: np.ndarray
    _periodic: List[bool]
    _shape: Tuple[int, ...]

    # to help sphinx, we here list docstrings for classproperties
    operators: Set[str]
    """ set: names of all operators defined for this grid """

    def __init__(self) -> None:
        """initialize the grid"""
        self._logger = logging.getLogger(self.__class__.__name__)
        self._mesh: Optional[GridMesh] = None

    def __init_subclass__(cls, **kwargs) -> None:  # @NoSelf
        """register all subclassess to reconstruct them later"""
        super().__init_subclass__(**kwargs)
        if cls is not GridBase:
            if cls.__name__ in cls._subclasses:
                warnings.warn(f"Redefining class {cls.__name__}")
            cls._subclasses[cls.__name__] = cls
        cls._operators = {}

    @classmethod
    def from_state(cls, state: Union[str, Dict[str, Any]]) -> GridBase:
        """create a field from a stored `state`.

        Args:
            state (`str` or `dict`):
                The state from which the grid is reconstructed. If `state` is a
                string, it is decoded as JSON, which should yield a `dict`.

        Returns:
            :class:`GridBase`: Grid re-created from the state
        """
        # decode the json data
        if isinstance(state, str):
            state = dict(json.loads(state))

        # create the instance of the correct class
        class_name = state.pop("class")
        if class_name == cls.__name__:
            raise RuntimeError(f"Cannot reconstruct abstract class `{class_name}`")
        grid_cls = cls._subclasses[class_name]
        return grid_cls.from_state(state)

    @classmethod
    def from_bounds(
        cls,
        bounds: Sequence[Tuple[float, float]],
        shape: Sequence[int],
        periodic: Sequence[bool],
    ) -> GridBase:
        raise NotImplementedError

    @property
    def periodic(self) -> List[bool]:
        """list: Flags that describe which axes are periodic"""
        return self._periodic

    @property
    def axes_bounds(self) -> Tuple[Tuple[float, float], ...]:
        """tuple: lower and upper bounds of each axis"""
        return self._axes_bounds

    @property
    def axes_coords(self) -> Tuple[np.ndarray, ...]:
        """tuple: coordinates of the cells for each axis"""
        return self._axes_coords

    def get_axis_index(self, key: Union[int, str], allow_symmetric: bool = True) -> int:
        """return the index belonging to an axis

        Args:
            key (int or str):
                The index or name of an axis
            allow_symmetric (bool):
                Whether axes with assumed symmetry are included

        Returns:
            int: The index of the axis
        """
        if isinstance(key, str):
            # determine key index from name of the axis
            if allow_symmetric:
                axes = self.axes + self.axes_symmetric
            else:
                axes = self.axes
            if key in axes:
                return axes.index(key)
            else:
                raise IndexError(f"`{key}` is not in the axes {axes}")
        elif isinstance(key, int):
            # assume that it is already an index
            return key
        raise IndexError("Index must be an integer or the name of an axes")

    def _get_boundary_index(
        self, index: Union[str, Tuple[int, bool]]
    ) -> Tuple[int, bool]:
        """return the index of a boundary belonging to an axis

        Args:
            index (str or tuple):
                Index specifying the boundary. Can be either a string given in
                :attr:`~pde.grids.base.GridBase.boundary_names`, like :code:`"left"`, or
                a tuple of the axis index perpendicular to the boundary and a boolean
                specifying whether the boundary is at the upper side of the axis or not,
                e.g., :code:`(1, True)`.

        Returns:
            tuple: axis index perpendicular to the boundary and a boolean specifying
                whether the boundary is at the upper side of the axis or not.
        """
        if isinstance(index, str):
            # assume that the index is a known identifier
            axis, upper = self.boundary_names[index]
        else:
            axis, upper = index
        return axis, upper

    @property
    def discretization(self) -> np.ndarray:
        """:class:`numpy.array`: the linear size of a cell along each axis"""
        return self._discretization

    @property
    def shape(self) -> Tuple[int, ...]:
        """tuple of int: the number of support points of each axis"""
        return self._shape

    @property
    def num_cells(self) -> int:
        """int: the number of cells in this grid"""
        return math.prod(self.shape)

    @property
    def _shape_full(self) -> Tuple[int, ...]:
        """tuple of int: number of support points including ghost points"""
        return tuple(num + 2 for num in self.shape)

    @property
    def _idx_valid(self) -> Tuple[slice, ...]:
        """tuple: slices to extract valid data from full data"""
        return tuple(slice(1, s + 1) for s in self.shape)

    def _make_get_valid(self) -> Callable[[np.ndarray], np.ndarray]:
        """create a function to extract the valid part of a full data array

        Returns:
            callable: Mapping a numpy array containing the full data of the grid to a
                numpy array of only the valid data
        """
        num_axes = self.num_axes

        @jit
        def get_valid(arr: np.ndarray) -> np.ndarray:
            """return valid part of the data (without ghost cells)"""
            if num_axes == 1:
                return arr[..., 1:-1]
            elif num_axes == 2:
                return arr[..., 1:-1, 1:-1]
            elif num_axes == 3:
                return arr[..., 1:-1, 1:-1, 1:-1]
            else:
                raise NotImplementedError

        return get_valid  # type: ignore

    def _make_set_valid(self) -> Callable[[np.ndarray, np.ndarray], None]:
        """create a function to set the valid part of a full data array

        Returns:
            callable: Takes two numpy arrays, setting the valid data in the first one,
                using the second array
        """
        num_axes = self.num_axes

        @jit
        def set_valid(arr: np.ndarray, value: np.ndarray) -> None:
            """return valid part of the data (without ghost cells)"""
            if num_axes == 1:
                arr[..., 1:-1] = value
            elif num_axes == 2:
                arr[..., 1:-1, 1:-1] = value
            elif num_axes == 3:
                arr[..., 1:-1, 1:-1, 1:-1] = value
            else:
                raise NotImplementedError

        return set_valid  # type: ignore

    @property
    @abstractmethod
    def state(self) -> Dict[str, Any]:
        ...

    @property
    def state_serialized(self) -> str:
        """str: JSON-serialized version of the state of this grid"""
        state = self.state
        state["class"] = self.__class__.__name__
        return json.dumps(state)

    def copy(self) -> GridBase:
        """return a copy of the grid"""
        return self.__class__.from_state(self.state)

    __copy__ = copy

    def __deepcopy__(self, memo: Dict[int, Any]) -> GridBase:
        """create a deep copy of the grid. This function is for instance called when
        a grid instance appears in another object that is copied using `copy.deepcopy`
        """
        # this implementation assumes that a simple call to copy is sufficient
        result = self.copy()
        memo[id(self)] = result
        return result

    def __repr__(self) -> str:
        """return instance as string"""
        args = ", ".join(str(k) + "=" + str(v) for k, v in self.state.items())
        return f"{self.__class__.__name__}({args})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.shape == other.shape
            and self.axes_bounds == other.axes_bounds
            and self.periodic == other.periodic
        )

    def _cache_hash(self) -> int:
        """returns a value to determine when a cache needs to be updated"""
        return hash(
            (
                self.__class__.__name__,
                self.shape,
                self.axes_bounds,
                tuple(self.periodic),
            )
        )

    def compatible_with(self, other: GridBase) -> bool:
        """tests whether this grid is compatible with other grids.

        Grids are compatible when they cover the same area with the same
        discretization. The difference to equality is that compatible grids do
        not need to have the same periodicity in their boundaries.

        Args:
            other (:class:`~pde.grids.base.GridBase`):
                The other grid to test against

        Returns:
            bool: Whether the grid is compatible
        """
        return (
            self.__class__ == other.__class__
            and self.shape == other.shape
            and self.axes_bounds == other.axes_bounds
        )

    def assert_grid_compatible(self, other: GridBase) -> None:
        """checks whether `other` is compatible with the current grid

        Args:
            other (:class:`~pde.grids.base.GridBase`):
                The grid compared to this one

        Raises:
            ValueError: if grids are not compatible
        """
        if not self.compatible_with(other):
            raise ValueError(f"Grids {self} and {other} are incompatible")

    @property
    def numba_type(self) -> str:
        """str: represents type of the grid data in numba signatures"""
        return "f8[" + ", ".join([":"] * self.num_axes) + "]"

    @cached_property()
    def coordinate_arrays(self) -> Tuple[np.ndarray, ...]:
        """tuple: for each axes: coordinate values for all cells"""
        return tuple(np.meshgrid(*self.axes_coords, indexing="ij"))

    @cached_property()
    def cell_coords(self) -> np.ndarray:
        """:class:`~numpy.ndarray`: coordinate values for all axes of each cell"""
        return np.moveaxis(self.coordinate_arrays, 0, -1)

    @cached_property()
    def cell_volumes(self) -> np.ndarray:
        """:class:`~numpy.ndarray`: volume of each cell"""
        vols = functools.reduce(np.outer, self.cell_volume_data)
        return np.broadcast_to(vols, self.shape)

    @cached_property()
    def uniform_cell_volumes(self) -> bool:
        """bool: returns True if all cell volumes are the same"""
        return all(np.asarray(vols).ndim == 0 for vols in self.cell_volume_data)

    def difference_vector_real(self, p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
        """return vector(s) pointing from p1 to p2

        In case of periodic boundary conditions, the shortest vector is returned.

        Warning:
            This function assumes a Cartesian coordinate system and might thus not work
            for curvilinear coordinates.

        Args:
            p1 (:class:`~numpy.ndarray`):
                First point(s)
            p2 (:class:`~numpy.ndarray`):
                Second point(s)

        Returns:
            :class:`~numpy.ndarray`: The difference vectors between the points with
            periodic  boundary conditions applied.
        """
        diff = np.atleast_1d(p2) - np.atleast_1d(p1)
        assert diff.shape[-1] == self.num_axes

        for i, periodic in enumerate(self.periodic):
            if periodic:
                size = self.axes_bounds[i][1] - self.axes_bounds[i][0]
                diff[..., i] = (diff[..., i] + size / 2) % size - size / 2
        return diff  # type: ignore

    def distance_real(self, p1: np.ndarray, p2: np.ndarray) -> float:
        """Calculate the distance between two points given in real coordinates

        This takes periodic boundary conditions into account if necessary.

        Warning:
            This function calculates the Euclidean distance and might thus give wrong
            results for coordinates given in curvilinear coordinate systems.

        Args:
            p1 (:class:`~numpy.ndarray`):
                First position
            p2 (:class:`~numpy.ndarray`):
                Second position

        Returns:
            float: Distance between the two positions
        """
        diff = self.difference_vector_real(p1, p2)
        return np.linalg.norm(diff, axis=-1)  # type: ignore

    def _iter_boundaries(self) -> Iterator[Tuple[int, bool]]:
        """iterate over all boundaries of the grid

        Yields:
            tuple: for each boundary, the generator returns a tuple indicating
            the axis of the boundary together with a boolean value indicating
            whether the boundary lies on the upper side of the axis.
        """
        return itertools.product(range(self.num_axes), [True, False])

    def _boundary_coordinates(self, axis: int, upper: bool) -> np.ndarray:
        """get coordinates of points on the boundary

        Args:
            axis (int):
                The axis perpendicular to the boundary
            upper (bool):
                Whether the boundary is at the upper side of the axis

        Returns:
            :class:`~numpy.ndarray`: Coordinates of the boundary points. This array has
            one less dimension than the grid has axes.
        """
        # get coordinate along the axis determining the boundary
        if upper:
            c_bndry = np.array([self._axes_bounds[axis][1]])
        else:
            c_bndry = np.array([self._axes_bounds[axis][0]])

        # get orthogonal coordinates
        coords = tuple(
            c_bndry if i == axis else self._axes_coords[i] for i in range(self.num_axes)
        )
        points = np.meshgrid(*coords, indexing="ij")

        # assemble into array
        shape_bndry = tuple(self.shape[i] for i in range(self.num_axes) if i != axis)
        shape = shape_bndry + (self.num_axes,)
        return np.stack(points, -1).reshape(shape)  # type: ignore

    @property
    @abstractmethod
    def volume(self) -> float:
        ...

    @abstractmethod
    def point_to_cartesian(
        self, points: np.ndarray, *, full: bool = False
    ) -> np.ndarray:
        ...

    @abstractmethod
    def point_from_cartesian(self, points: np.ndarray) -> np.ndarray:
        ...

    def normalize_point(
        self, point: np.ndarray, *, reflect: bool = False
    ) -> np.ndarray:
        """normalize grid coordinates by applying periodic boundary conditions

        Here, points are assumed to be specified by the physical values along the
        non-symmetric axes of the grid, e.g., by grid coordinates. Normalizing points is
        useful to make sure they lie within the domain of the  grid. This function
        respects periodic boundary conditions and can also reflect points off the
        boundary if `reflect = True`.

        Args:
            point (:class:`~numpy.ndarray`):
                Coordinates of a single point or an array of points, where the last axis
                denotes the point coordinates (e.g., a list of points).
            reflect (bool):
                Flag determining whether coordinates along non-periodic axes are
                reflected to lie in the valid range. If `False`, such coordinates are
                left unchanged and only periodic boundary conditions are enforced.

        Returns:
            :class:`~numpy.ndarray`: The respective coordinates with periodic
            boundary conditions applied.
        """
        point = np.asarray(point, dtype=np.double)
        if point.size == 0:
            return np.zeros((0, self.num_axes))

        if point.ndim == 0:
            if self.num_axes > 1:
                raise DimensionError(
                    f"Point {point} is not of dimension {self.num_axes}"
                )
        elif point.shape[-1] != self.num_axes:
            raise DimensionError(
                f"Array of shape {point.shape} does not describe points of dimension "
                f"{self.num_axes}"
            )

        # normalize the coordinates for the periodic dimensions
        bounds = np.array(self.axes_bounds)
        xmin = bounds[:, 0]
        xmax = bounds[:, 1]
        xdim = xmax - xmin

        if self.num_axes == 1:
            # single dimension
            if self.periodic[0]:
                point = (point - xmin[0]) % xdim[0] + xmin[0]
            elif reflect:
                arg = (point - xmax[0]) % (2 * xdim[0]) - xdim[0]
                point = xmin[0] + np.abs(arg)

        else:
            # multiple dimensions
            for i in range(self.num_axes):
                if self.periodic[i]:
                    point[..., i] = (point[..., i] - xmin[i]) % xdim[i] + xmin[i]
                elif reflect:
                    arg = (point[..., i] - xmax[i]) % (2 * xdim[i]) - xdim[i]
                    point[..., i] = xmin[i] + np.abs(arg)

        return point

    def _grid_to_cell(
        self, grid_coords: np.ndarray, *, truncate: bool = True, normalize: bool = False
    ) -> np.ndarray:
        """convert grid coordinates to cell coordinates

        Args:
            grid_coords (:class:`~numpy.ndarray`):
                The grid coordinates to convert
            truncate (bool):
                Flag indicating whether the resulting cell coordinates are integers
                marking what cell the point belongs to or whether fractional coordinates
                are returned. The default is to return integers.
            normalize (bool):
                Flag indicating whether the points should be normalized, e.g., whether
                periodic boundary conditions should be imposed.

        Returns:
            :class:`~numpy.ndarray`: The cell coordinates
        """
        if normalize:
            coords = self.normalize_point(grid_coords)
        else:
            coords = np.atleast_1d(grid_coords)

        # convert from grid coordinates to cells indices
        c_min = np.array(self.axes_bounds)[:, 0]
        cells = (coords - c_min) / self.discretization
        if truncate:
            return cells.astype(int)  # type: ignore
        else:
            return cells  # type: ignore

    def transform(
        self,
        coordinates: np.ndarray,
        source: Literal["cartesian", "cell", "grid"],
        target: Literal["cartesian", "cell", "grid"],
    ) -> np.ndarray:
        """converts coordinates from one coordinate system to another

        Supported coordinate systems include the following:

        * `cartesian`:
            Cartesian coordinates where each point carries `dim` values. These are the
            true physical coordinates in space.
        * `grid`:
            Coordinates values in the coordinate system defined by the grid. A point is
            thus characterized by `grid.num_axes` values.
        * `cell`:
            Normalized grid coordinates based on indexing the discretization cells. A
            point is characterized by `grid.num_axes` values and the range of values for
            a given axis is between `0` and `N`, where `N` is the number of grid points.
            Consequently, the integral part of the cell coordinate denotes the cell,
            while the fractional part denotes the relative position within the cell. In
            particular, the cell center is located at `i + 0.5` with `i = 0, ..., N-1`.

        Note:
            Some conversion might involve projections if the coordinate system imposes
            symmetries. For instance, converting 3d Cartesian coordinates to grid
            coordinates in a spherically symmetric  grid will only return the radius
            from the origin. Conversely, converting these grid coordinates back to 3d
            Cartesian coordinates will only return coordinates along a particular ray
            originating at the origin.

        Args:
            coordinates (:class:`~numpy.ndarray`):
                The coordinates to convert
            source (str):
                The source coordinate system
            target (str):
                The target coordinate system

        Returns:
            :class:`~numpy.ndarray`: The transformed coordinates
        """
        if source == "cartesian":
            # Cartesian coordinates given
            cartesian = np.atleast_1d(coordinates)
            if cartesian.shape[-1] != self.dim:
                raise DimensionError(f"Require {self.dim} cartesian coordinates")

            if target == "cartesian":
                return coordinates

            # convert Cartesian coordinates to grid coordinates
            grid_coords = self.point_from_cartesian(cartesian)

            if target == "grid":
                return grid_coords
            if target == "cell":
                return self._grid_to_cell(grid_coords, truncate=False, normalize=False)

        elif source == "cell":
            # Cell coordinates given
            cells = np.atleast_1d(coordinates)
            if cells.shape[-1] != self.num_axes:
                raise DimensionError(f"Require {self.num_axes} cell coordinates")

            if target == "cell":
                return coordinates

            # convert cell coordinates to grid coordinates
            c_min = np.array(self.axes_bounds)[:, 0]
            grid_coords = c_min + cells * self.discretization

            if target == "grid":
                return grid_coords
            elif target == "cartesian":
                return self.point_to_cartesian(grid_coords, full=False)

        elif source == "grid":
            # Grid coordinates given
            grid_coords = np.atleast_1d(coordinates)
            if grid_coords.shape[-1] != self.num_axes:
                raise DimensionError(f"Require {self.num_axes} grid coordinates")

            if target == "cartesian":
                return self.point_to_cartesian(grid_coords, full=False)
            elif target == "cell":
                return self._grid_to_cell(grid_coords, truncate=False, normalize=False)
            elif target == "grid":
                return grid_coords

        else:
            raise ValueError(f"Unknown source coordinates `{source}`")
        raise ValueError(f"Unknown target coordinates `{target}`")

    @abstractmethod
    def polar_coordinates_real(
        self, origin: np.ndarray, *, ret_angle: bool = False
    ) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        ...

    def contains_point(
        self,
        points: np.ndarray,
        *,
        coords: Literal["cartesian", "cell", "grid"] = "cartesian",
    ) -> np.ndarray:
        """check whether the point is contained in the grid

        Args:
            point (:class:`~numpy.ndarray`):
                Coordinates of the point
            coords (str):
                The coordinate system in which the points are given

        Returns:
            :class:`~numpy.ndarray`: A boolean array indicating which points lie within
            the grid
        """
        cell_coords = self.transform(points, source=coords, target="cell")
        return np.all((0 <= cell_coords) & (cell_coords <= self.shape), axis=-1)  # type: ignore

    @abstractmethod
    def iter_mirror_points(
        self, point: np.ndarray, with_self: bool = False, only_periodic: bool = True
    ) -> Generator:
        ...

    @fill_in_docstring
    def get_boundary_conditions(
        self, bc: "BoundariesData" = "auto_periodic_neumann", rank: int = 0
    ) -> Boundaries:
        """constructs boundary conditions from a flexible data format

        Args:
            bc (str or list or tuple or dict):
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            rank (int):
                The tensorial rank of the value associated with the boundary conditions.

        Returns:
            :class:`~pde.grids.boundaries.axes.Boundaries`: The boundary conditions for
            all axes.

        Raises:
            ValueError:
                If the data given in `bc` cannot be read
            PeriodicityError:
                If the boundaries are not compatible with the periodic axes of the grid.
        """
        from .boundaries import Boundaries  # @Reimport

        if self._mesh is None:
            # get boundary conditions for a simple grid that is not part of a mesh
            bcs = Boundaries.from_data(self, bc, rank=rank)

        else:
            # this grid is part of a mesh and we thus need to set special conditions to
            # support parallelism via MPI. We here assume that bc is given for the full
            # system and not
            bcs_base = Boundaries.from_data(self._mesh.basegrid, bc, rank=rank)
            bcs = self._mesh.extract_boundary_conditions(bcs_base)

        return bcs

    @abstractmethod
    def get_line_data(self, data: np.ndarray, extract: str = "auto") -> Dict[str, Any]:
        ...

    @abstractmethod
    def get_image_data(self, data: np.ndarray) -> Dict[str, Any]:
        ...

    @abstractmethod
    def get_random_point(
        self, *, boundary_distance: float = 0, coords: CoordsType = "cartesian"
    ) -> np.ndarray:
        ...

    @classmethod
    def register_operator(
        cls,
        name: str,
        factory_func: Optional[OperatorFactory] = None,
        rank_in: int = 0,
        rank_out: int = 0,
    ):
        """register an operator for this grid

        Example:
            The method can either be used directly:

            .. code-block:: python

                GridClass.register_operator("operator", make_operator)

            or as a decorator for the factory function:

            .. code-block:: python

                @GridClass.register_operator("operator")
                def make_operator(grid: GridBase):
                    ...

        Args:
            name (str):
                The name of the operator to register
            factory_func (callable):
                A function with signature ``(grid: GridBase, **kwargs)``, which takes
                a grid object and optional keyword arguments and returns an
                implementation of the given operator. This implementation is a function
                that takes a :class:`~numpy.ndarray` of discretized values as arguments
                and returns the resulting discretized data in a :class:`~numpy.ndarray`
                after applying the operator.
            rank_in (int):
                The rank of the input field for the operator
            rank_out (int):
                The rank of the field that is returned by the operator
        """

        def register_operator(factor_func_arg: OperatorFactory):
            """helper function to register the operator"""
            cls._operators[name] = OperatorInfo(
                factory=factor_func_arg, rank_in=rank_in, rank_out=rank_out, name=name
            )
            return factor_func_arg

        if factory_func is None:
            # method is used as a decorator, so return the helper function
            return register_operator
        else:
            # method is used directly
            register_operator(factory_func)

    @hybridmethod  # type: ignore
    @property
    def operators(cls) -> Set[str]:  # @NoSelf
        """set: all operators defined for this class"""
        result = set()
        # add all customly defined operators
        classes = inspect.getmro(cls)[:-1]  # type: ignore
        for anycls in classes:
            result |= set(anycls._operators.keys())  # type: ignore
        if hasattr(cls, "axes"):
            for ax in cls.axes:
                result |= {f"d_d{ax}", f"d2_d{ax}2"}
        return result

    @operators.instancemethod
    @property
    def operators(self) -> Set[str]:
        """set: all operators defined for this instance"""
        # get all operators registered on the class
        result = self.__class__.operators
        if not hasattr(self.__class__, "axes"):
            # add operators calculating derivate along a coordinate for the case where
            # the axes argument is only defined on instances
            for ax in self.axes:
                result |= {f"d_d{ax}", f"d2_d{ax}2"}
        return result

    def _get_operator_info(self, operator: Union[str, OperatorInfo]) -> OperatorInfo:
        """return the operator defined on this grid

        Args:
            operator (str):
                Identifier for the operator. Some examples are 'laplace', 'gradient', or
                'divergence'. The registered operators for this grid can be obtained
                from the :attr:`~pde.grids.base.GridBase.operators` attribute.

        Returns:
            :class:`~pde.grids.base.OperatorInfo`: information for the operator
        """
        if isinstance(operator, OperatorInfo):
            return operator
        assert isinstance(operator, str)

        # look for defined operators on all parent classes (except `object`)
        classes = inspect.getmro(self.__class__)[:-1]
        for cls in classes:
            if operator in cls._operators:  # type: ignore
                return cls._operators[operator]  # type: ignore

        # deal with some special patterns that are often used
        if operator.startswith("d_d"):
            # create a special operator that takes a first derivative along one axis
            from .operators.cartesian import _make_derivative

            axis_id = self.axes.index(operator[len("d_d") :])
            factory = functools.partial(_make_derivative, axis=axis_id)
            return OperatorInfo(factory, rank_in=0, rank_out=0, name=operator)

        elif operator.startswith("d2_d") and operator.endswith("2"):
            # create a special operator that takes a second derivative along one axis
            from .operators.cartesian import _make_derivative2

            axis_id = self.axes.index(operator[len("d2_d") : -1])
            factory = functools.partial(_make_derivative2, axis=axis_id)
            return OperatorInfo(factory, rank_in=0, rank_out=0, name=operator)

        # throw an informative error since operator was not found
        op_list = ", ".join(sorted(self.operators))
        raise ValueError(
            f"'{operator}' is not one of the defined operators ({op_list}). Custom "
            "operators can be added using the `register_operator` method."
        )

    @cached_method()
    def make_operator_no_bc(
        self,
        operator: Union[str, OperatorInfo],
        **kwargs,
    ) -> OperatorType:
        """return a compiled function applying an operator without boundary conditions

        A function that takes the discretized full data as an input and an array of
        valid data points to which the result of applying the operator is written.

        Note:
            The resulting function does not check whether the ghost cells of the input
            array have been supplied with sensible values. It is the responsibility of
            the user to set the values of the ghost cells beforehand. Use this function
            only if you absolutely know what you're doing. In all other cases,
            :meth:`make_operator` is probably the better choice.

        Args:
            operator (str):
                Identifier for the operator. Some examples are 'laplace', 'gradient', or
                'divergence'. The registered operators for this grid can be obtained
                from the :attr:`~pde.grids.base.GridBase.operators` attribute.
            **kwargs:
                Specifies extra arguments influencing how the operator is created.

        Returns:
            callable: the function that applies the operator. This function has the
            signature (arr: np.ndarray, out: np.ndarray), so they `out` array need to be
            supplied explicitly.
        """
        return self._get_operator_info(operator).factory(self, **kwargs)

    @cached_method()
    @fill_in_docstring
    def make_operator(
        self, operator: Union[str, OperatorInfo], bc: BoundariesData, **kwargs
    ) -> Callable[..., np.ndarray]:
        """return a compiled function applying an operator with boundary conditions

        The returned function takes the discretized data on the grid as an input and
        returns the data to which the operator `operator` has been applied. The function
        only takes the valid grid points and allocates memory for the ghost points
        internally to apply the boundary conditions specified as `bc`. Note that the
        function supports an optional argument `out`, which if given should provide
        space for the valid output array without the ghost cells. The result of the
        operator is then written into this output array. The function also accepts an
        optional parameter `args`, which is forwarded to `set_ghost_cells`.

        Args:
            operator (str):
                Identifier for the operator. Some examples are 'laplace', 'gradient', or
                'divergence'. The registered operators for this grid can be obtained
                from the :attr:`~pde.grids.base.GridBase.operators` attribute.
            bc (str or list or tuple or dict):
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            **kwargs:
                Specifies extra arguments influencing how the operator is created.

        Returns:
            callable: the function that applies the operator. This function has the
            signature (arr: np.ndarray, out: np.ndarray = None, args=None).
        """
        backend = kwargs.get("backend", "numba")  # numba is the default backend

        # instantiate the operator
        operator = self._get_operator_info(operator)
        operator_raw = operator.factory(self, **kwargs)

        # set the boundary conditions before applying this operator
        bcs = self.get_boundary_conditions(bc, rank=operator.rank_in)

        # calculate shapes of the full data
        shape_in_valid = (self.dim,) * operator.rank_in + self.shape
        shape_in_full = (self.dim,) * operator.rank_in + self._shape_full
        shape_out = (self.dim,) * operator.rank_out + self.shape

        # define numpy version of the operator
        def apply_op(
            arr: np.ndarray, out: Optional[np.ndarray] = None, args=None
        ) -> np.ndarray:
            """set boundary conditions and apply operator"""
            assert arr.shape == shape_in_valid
            # ensure `out` array is allocated
            if out is None:
                out = np.empty(shape_out, dtype=arr.dtype)
            else:
                assert out.shape == shape_out

            # prepare input with boundary conditions
            arr_full = np.empty(shape_in_full, dtype=arr.dtype)
            arr_full[(...,) + self._idx_valid] = arr
            bcs.set_ghost_cells(arr_full, args=args)

            # apply operator
            operator_raw(arr_full, out)

            # return valid part of the output
            return out

        if backend in {"numpy", "scipy"}:
            # return the bare operator without the numba-overloaded version
            return apply_op

        elif backend.startswith("numba"):
            # overload `apply_op` with numba-compiled version
            set_ghost_cells = bcs.make_ghost_cell_setter()
            set_valid = self._make_set_valid()

            if not is_jitted(operator_raw):
                operator_raw = jit(operator_raw)

            @overload(apply_op, inline="always")
            def apply_op_ol(
                arr: np.ndarray, out: Optional[np.ndarray] = None, args=None
            ) -> np.ndarray:
                """make numba implementation of the operator"""
                if isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                    # need to allocate memory for `out`

                    def apply_op_impl(
                        arr: np.ndarray, out: Optional[np.ndarray] = None, args=None
                    ) -> np.ndarray:
                        """allocates `out` and applies operator to the data"""
                        assert arr.shape == shape_in_valid

                        out = np.empty(shape_out, dtype=arr.dtype)
                        # prepare input with boundary conditions
                        arr_full = np.empty(shape_in_full, dtype=arr.dtype)
                        set_valid(arr_full, arr)
                        set_ghost_cells(arr_full, args=args)

                        # apply operator
                        operator_raw(arr_full, out)

                        # return valid part of the output
                        return out

                else:
                    # reuse provided `out` array

                    def apply_op_impl(
                        arr: np.ndarray, out: Optional[np.ndarray] = None, args=None
                    ) -> np.ndarray:
                        """applies operator to the data wihtout allocating out"""
                        assert arr.shape == shape_in_valid
                        assert out.shape == shape_out  # type: ignore

                        # prepare input with boundary conditions
                        arr_full = np.empty(shape_in_full, dtype=arr.dtype)
                        set_valid(arr_full, arr)
                        set_ghost_cells(arr_full, args=args)

                        # apply operator
                        operator_raw(arr_full, out)  # type: ignore

                        # return valid part of the output
                        return out  # type: ignore

                return apply_op_impl  # type: ignore

            @jit
            def apply_op_compiled(
                arr: np.ndarray, out: Optional[np.ndarray] = None, args=None
            ) -> np.ndarray:
                """set boundary conditions and apply operator"""
                return apply_op(arr, out, args)

            # return the compiled versions of the operator
            return apply_op_compiled  # type: ignore

        else:
            # simply return the operator if the backend was `numba` or `scipy`
            raise NotImplementedError(f"Undefined backend '{backend}'")

    def slice(self, indices: Sequence[int]) -> GridBase:
        """return a subgrid of only the specified axes

        Args:
            indices (list):
                Indices indicating the axes that are retained in the subgrid

        Returns:
            :class:`GridBase`: The subgrid
        """
        raise NotImplementedError(
            f"Slicing is not implemented for class {self.__class__.__name__}"
        )

    def plot(self) -> None:
        """visualize the grid"""
        raise NotImplementedError(
            f"Plotting is not implemented for class {self.__class__.__name__}"
        )

    @property
    def typical_discretization(self) -> float:
        """float: the average side length of the cells"""
        return np.mean(self.discretization)  # type: ignore

    def integrate(
        self, data: NumberOrArray, axes: Union[int, Sequence[int], None] = None
    ) -> NumberOrArray:
        """Integrates the discretized data over the grid

        Args:
            data (:class:`~numpy.ndarray`):
                The values at the support points of the grid that need to be
                integrated.
            axes (list of int, optional):
                The axes along which the integral is performed. If omitted, all
                axes are integrated over.

        Returns:
            :class:`~numpy.ndarray`: The values integrated over the entire grid
        """
        # determine the volumes of the individual cells
        if axes is None:
            volume_list = self.cell_volume_data
        else:
            # use stored value for the default case of integrating over all axes
            if isinstance(axes, int):
                axes = (axes,)
            else:
                axes = tuple(axes)  # required for numpy.sum
            volume_list = [
                cell_vol if ax in axes else 1
                for ax, cell_vol in enumerate(self.cell_volume_data)
            ]
        cell_volumes = functools.reduce(np.outer, volume_list)

        # determine the axes over which we will integrate
        if not isinstance(data, np.ndarray) or data.ndim < self.num_axes:
            # deal with the case where data is not supplied for each support
            # point, e.g., when a single scalar is integrated over the grid
            data = np.broadcast_to(data, self.shape)

        elif data.ndim > self.num_axes:
            # deal with the case where more than a single value is provided per
            # support point, e.g., when a tensorial field is integrated
            offset = data.ndim - self.num_axes
            if axes is None:
                # integrate over all axes of the grid
                axes = tuple(range(offset, data.ndim))
            else:
                # shift the indices to account for the data shape
                axes = tuple(offset + i for i in axes)

        # calculate integral using a weighted sum along the chosen axes
        integral = (data * cell_volumes).sum(axis=axes)

        if self._mesh is None or len(self._mesh) == 1:
            # standard case of a single integral
            return integral  # type: ignore

        else:
            # we are in a parallel run, so we need to gather the sub-integrals from all
            from mpi4py.MPI import COMM_WORLD

            integral_full = np.empty_like(integral)
            COMM_WORLD.Allreduce(integral, integral_full)
            return integral_full  # type: ignore

    @cached_method()
    def make_normalize_point_compiled(
        self, reflect: bool = True
    ) -> Callable[[np.ndarray], None]:
        """return a compiled function that normalizes a point

        Here, the point is assumed to be specified by the physical values along
        the non-symmetric axes of the grid. Normalizing points is useful to make sure
        they lie within the domain of the  grid. This function respects periodic
        boundary conditions and can also reflect points off the boundary.

        Args:
            reflect (bool):
                Flag determining whether coordinates along non-periodic axes are
                reflected to lie in the valid range. If `False`, such coordinates are
                left unchanged and only periodic boundary conditions are enforced.

        Returns:
            callable: A function that takes a :class:`~numpy.ndarray` as an argument,
            which describes the coordinates of the points. This array is modified
            in-place!
        """
        num_axes = self.num_axes
        periodic = np.array(self.periodic)  # using a tuple instead led to a numba error
        bounds = np.array(self.axes_bounds)
        xmin = bounds[:, 0]
        xmax = bounds[:, 1]
        size = bounds[:, 1] - bounds[:, 0]

        @jit
        def normalize_point(point: np.ndarray) -> None:
            """helper function normalizing a single point"""
            assert point.ndim == 1  # only support single points
            for i in range(num_axes):
                if periodic[i]:
                    point[i] = (point[i] - xmin[i]) % size[i] + xmin[i]
                elif reflect:
                    arg = (point[i] - xmax[i]) % (2 * size[i]) - size[i]
                    point[i] = xmin[i] + abs(arg)
                # else: do nothing

        return normalize_point  # type: ignore

    @cached_method()
    def make_cell_volume_compiled(self, flat_index: bool = False) -> CellVolume:
        """return a compiled function returning the volume of a grid cell

        Args:
            flat_index (bool):
                When True, cell_volumes are indexed by a single integer into the
                flattened array.

        Returns:
            function: returning the volume of the chosen cell
        """
        if all(np.isscalar(d) for d in self.cell_volume_data):
            # all cells have the same volume
            cell_volume = np.prod(self.cell_volume_data)  # type: ignore

            @jit
            def get_cell_volume(*args) -> float:
                return cell_volume  # type: ignore

        else:
            # some cells have a different volume
            cell_volumes = self.cell_volumes

            if flat_index:

                @jit
                def get_cell_volume(idx: int) -> float:
                    return cell_volumes.flat[idx]  # type: ignore

            else:

                @jit
                def get_cell_volume(*args) -> float:
                    return cell_volumes[args]  # type: ignore

        return get_cell_volume  # type: ignore

    def _make_interpolation_axis_data(
        self,
        axis: int,
        *,
        with_ghost_cells: bool = False,
        cell_coords: bool = False,
    ) -> Callable[[float], Tuple[int, int, float, float]]:
        """factory for obtaining interpolation information

        Args:
            axis (int):
                The axis along which interpolation is performed
            with_ghost_cells (bool):
                Flag indicating that the interpolator should work on the full data array
                that includes values for the ghost points. If this is the case, the
                boundaries are not checked and the coordinates are used as is.
            cell_coords (bool):
                Flag indicating whether points are given in cell coordinates or actual
                point coordinates.

        Returns:
            callable: A function that is called with a coordinate value for the axis.
            The function returns the indices of the neighboring support points as well
            as the associated weights.
        """
        # obtain information on how this axis is discretized
        size = self.shape[axis]
        periodic = self.periodic[axis]
        lo = self.axes_bounds[axis][0]
        dx = self.discretization[axis]

        @register_jitable
        def get_axis_data(coord: float) -> Tuple[int, int, float, float]:
            """determines data for interpolating along one axis"""
            # determine the index of the left cell and the fraction toward the right
            if cell_coords:
                c_l, d_l = divmod(coord, 1.0)
            else:
                c_l, d_l = divmod((coord - lo) / dx - 0.5, 1.0)

            # determine the indices of the two cells whose value affect interpolation
            if periodic:
                # deal with periodic domains, which is easy
                c_li = int(c_l) % size  # left support point
                c_hi = (c_li + 1) % size  # right support point

            elif with_ghost_cells:
                # deal with edge cases using the values of ghost cells
                if -0.5 <= c_l + d_l <= size - 0.5:  # in bulk part of domain
                    c_li = int(c_l)  # left support point
                    c_hi = c_li + 1  # right support point
                else:
                    return -42, -42, 0.0, 0.0  # indicates out of bounds

            else:
                # deal with edge cases using nearest-neighbor interpolation at boundary
                if 0 <= c_l + d_l < size - 1:  # in bulk part of domain
                    c_li = int(c_l)  # left support point
                    c_hi = c_li + 1  # right support point
                elif size - 1 <= c_l + d_l <= size - 0.5:  # close to upper boundary
                    c_li = c_hi = int(c_l)  # both support points close to boundary
                    # This branch also covers the special case, where size == 1 and data
                    # is evaluated at the only support point (c_l == d_l == 0.)
                elif -0.5 <= c_l + d_l <= 0:  # close to lower boundary
                    c_li = c_hi = int(c_l) + 1  # both support points close to boundary
                else:
                    return -42, -42, 0.0, 0.0  # indicates out of bounds

            # determine the weights of the two cells
            w_l, w_h = 1 - d_l, d_l
            # set small weights to zero. If this is not done, invalid data at the corner
            # of the grid (where two rows of ghost cells intersect) could be accessed.
            # If this random data is very large, e.g., 1e100, it contributes
            # significantly, even if the weight is low, e.g., 1e-16.
            if w_l < 1e-15:
                w_l = 0
            if w_h < 1e-15:
                w_h = 0

            # shift points to allow accessing data with ghost points
            if with_ghost_cells:
                c_li += 1
                c_hi += 1

            return c_li, c_hi, w_l, w_h

        return get_axis_data  # type: ignore

    @cached_method()
    def _make_interpolator_compiled(
        self,
        *,
        fill: Optional[Number] = None,
        with_ghost_cells: bool = False,
        cell_coords: bool = False,
    ) -> Callable[[np.ndarray, np.ndarray], np.ndarray]:
        """return a compiled function for linear interpolation on the grid

        Args:
            fill (Number, optional):
                Determines how values out of bounds are handled. If `None`, `ValueError`
                is raised when out-of-bounds points are requested. Otherwise, the given
                value is returned.
            with_ghost_cells (bool):
                Flag indicating that the interpolator should work on the full data array
                that includes values for the ghost points. If this is the case, the
                boundaries are not checked and the coordinates are used as is.
            cell_coords (bool):
                Flag indicating whether points are given in cell coordinates or actual
                point coordinates.

        Returns:
            callable: A function which returns interpolated values when called with
            arbitrary positions within the space of the grid. The signature of this
            function is (data, point), where `data` is the numpy array containing the
            field data and position denotes the position in grid coordinates.
        """
        args = {"with_ghost_cells": with_ghost_cells, "cell_coords": cell_coords}

        if self.num_axes == 1:
            # specialize for 1-dimensional interpolation
            data_x = self._make_interpolation_axis_data(0, **args)

            @jit
            def interpolate_single(
                data: np.ndarray, point: np.ndarray
            ) -> NumberOrArray:
                """obtain interpolated value of data at a point

                Args:
                    data (:class:`~numpy.ndarray`):
                        A 1d array of valid values at the grid points
                    point (:class:`~numpy.ndarray`):
                        Coordinates of a single point in the grid coordinate system

                Returns:
                    :class:`~numpy.ndarray`: The interpolated value at the point
                """
                c_li, c_hi, w_l, w_h = data_x(point[0])

                if c_li == -42:  # out of bounds
                    if fill is None:  # outside the domain
                        raise DomainError("Point lies outside the grid domain")
                    else:
                        return fill

                # do the linear interpolation
                return w_l * data[..., c_li] + w_h * data[..., c_hi]

        elif self.num_axes == 2:
            # specialize for 2-dimensional interpolation
            data_x = self._make_interpolation_axis_data(0, **args)
            data_y = self._make_interpolation_axis_data(1, **args)

            @jit
            def interpolate_single(
                data: np.ndarray, point: np.ndarray
            ) -> NumberOrArray:
                """obtain interpolated value of data at a point

                Args:
                    data (:class:`~numpy.ndarray`):
                        A 2d array of valid values at the grid points
                    point (:class:`~numpy.ndarray`):
                        Coordinates of a single point in the grid coordinate system

                Returns:
                    :class:`~numpy.ndarray`: The interpolated value at the point
                """
                # determine surrounding points and their weights
                c_xli, c_xhi, w_xl, w_xh = data_x(point[0])
                c_yli, c_yhi, w_yl, w_yh = data_y(point[1])

                if c_xli == -42 or c_yli == -42:  # out of bounds
                    if fill is None:  # outside the domain
                        raise DomainError("Point lies outside the grid domain")
                    else:
                        return fill

                # do the linear interpolation
                return (  # type: ignore
                    w_xl * w_yl * data[..., c_xli, c_yli]
                    + w_xl * w_yh * data[..., c_xli, c_yhi]
                    + w_xh * w_yl * data[..., c_xhi, c_yli]
                    + w_xh * w_yh * data[..., c_xhi, c_yhi]
                )

        elif self.num_axes == 3:
            # specialize for 3-dimensional interpolation
            data_x = self._make_interpolation_axis_data(0, **args)
            data_y = self._make_interpolation_axis_data(1, **args)
            data_z = self._make_interpolation_axis_data(2, **args)

            @jit
            def interpolate_single(
                data: np.ndarray, point: np.ndarray
            ) -> NumberOrArray:
                """obtain interpolated value of data at a point

                Args:
                    data (:class:`~numpy.ndarray`):
                        A 2d array of valid values at the grid points
                    point (:class:`~numpy.ndarray`):
                        Coordinates of a single point in the grid coordinate system

                Returns:
                    :class:`~numpy.ndarray`: The interpolated value at the point
                """
                # determine surrounding points and their weights
                c_xli, c_xhi, w_xl, w_xh = data_x(point[0])
                c_yli, c_yhi, w_yl, w_yh = data_y(point[1])
                c_zli, c_zhi, w_zl, w_zh = data_z(point[2])

                if c_xli == -42 or c_yli == -42 or c_zli == -42:  # out of bounds
                    if fill is None:  # outside the domain
                        raise DomainError("Point lies outside the grid domain")
                    else:
                        return fill

                # do the linear interpolation
                return (  # type: ignore
                    w_xl * w_yl * w_zl * data[..., c_xli, c_yli, c_zli]
                    + w_xl * w_yl * w_zh * data[..., c_xli, c_yli, c_zhi]
                    + w_xl * w_yh * w_zl * data[..., c_xli, c_yhi, c_zli]
                    + w_xl * w_yh * w_zh * data[..., c_xli, c_yhi, c_zhi]
                    + w_xh * w_yl * w_zl * data[..., c_xhi, c_yli, c_zli]
                    + w_xh * w_yl * w_zh * data[..., c_xhi, c_yli, c_zhi]
                    + w_xh * w_yh * w_zl * data[..., c_xhi, c_yhi, c_zli]
                    + w_xh * w_yh * w_zh * data[..., c_xhi, c_yhi, c_zhi]
                )

        else:
            raise NotImplementedError(
                f"Compiled interpolation not implemented for dimension {self.num_axes}"
            )

        return interpolate_single  # type: ignore

    def make_inserter_compiled(
        self, *, with_ghost_cells: bool = False
    ) -> Callable[[np.ndarray, np.ndarray, NumberOrArray], None]:
        """return a compiled function to insert values at interpolated positions

        Args:
            with_ghost_cells (bool):
                Flag indicating that the interpolator should work on the full data array
                that includes values for the grid points. If this is the case, the
                boundaries are not checked and the coordinates are used as is.

        Returns:
            callable: A function with signature (data, position, amount), where `data`
            is the numpy array containing the field data, position is denotes the
            position in grid coordinates, and `amount` is the  that is to be added to
            the field.
        """
        cell_volume = self.make_cell_volume_compiled()

        if self.num_axes == 1:
            # specialize for 1-dimensional interpolation
            data_x = self._make_interpolation_axis_data(
                0, with_ghost_cells=with_ghost_cells
            )

            @jit
            def insert(
                data: np.ndarray, point: np.ndarray, amount: NumberOrArray
            ) -> None:
                """add an amount to a field at an interpolated position

                Args:
                    data (:class:`~numpy.ndarray`):
                        The values at the grid points
                    point (:class:`~numpy.ndarray`):
                        Coordinates of a single point in the grid coordinate system
                    amount (Number or :class:`~numpy.ndarray`):
                        The amount that will be added to the data. This value describes
                        an integrated quantity (given by the field value times the
                        discretization volume). This is important for consistency with
                        different discretizations and in particular grids with
                        non-uniform discretizations
                """
                c_li, c_hi, w_l, w_h = data_x(point[0])

                if c_li == -42:  # out of bounds
                    raise DomainError("Point lies outside the grid domain")

                data[..., c_li] += w_l * amount / cell_volume(c_li)
                data[..., c_hi] += w_h * amount / cell_volume(c_hi)

        elif self.num_axes == 2:
            # specialize for 2-dimensional interpolation
            data_x = self._make_interpolation_axis_data(
                0, with_ghost_cells=with_ghost_cells
            )
            data_y = self._make_interpolation_axis_data(
                1, with_ghost_cells=with_ghost_cells
            )

            @jit
            def insert(
                data: np.ndarray, point: np.ndarray, amount: NumberOrArray
            ) -> None:
                """add an amount to a field at an interpolated position

                Args:
                    data (:class:`~numpy.ndarray`):
                        The values at the grid points
                    point (:class:`~numpy.ndarray`):
                        Coordinates of a single point in the grid coordinate system
                    amount (Number or :class:`~numpy.ndarray`):
                        The amount that will be added to the data. This value describes
                        an integrated quantity (given by the field value times the
                        discretization volume). This is important for consistency with
                        different discretizations and in particular grids with
                        non-uniform discretizations
                """
                # determine surrounding points and their weights
                c_xli, c_xhi, w_xl, w_xh = data_x(point[0])
                c_yli, c_yhi, w_yl, w_yh = data_y(point[1])

                if c_xli == -42 or c_yli == -42:  # out of bounds
                    raise DomainError("Point lies outside the grid domain")

                cell_vol = cell_volume(c_xli, c_yli)
                data[..., c_xli, c_yli] += w_xl * w_yl * amount / cell_vol
                cell_vol = cell_volume(c_xli, c_yhi)
                data[..., c_xli, c_yhi] += w_xl * w_yh * amount / cell_vol

                cell_vol = cell_volume(c_xhi, c_yli)
                data[..., c_xhi, c_yli] += w_xh * w_yl * amount / cell_vol
                cell_vol = cell_volume(c_xhi, c_yhi)
                data[..., c_xhi, c_yhi] += w_xh * w_yh * amount / cell_vol

        elif self.num_axes == 3:
            # specialize for 3-dimensional interpolation
            data_x = self._make_interpolation_axis_data(
                0, with_ghost_cells=with_ghost_cells
            )
            data_y = self._make_interpolation_axis_data(
                1, with_ghost_cells=with_ghost_cells
            )
            data_z = self._make_interpolation_axis_data(
                2, with_ghost_cells=with_ghost_cells
            )

            @jit
            def insert(
                data: np.ndarray, point: np.ndarray, amount: NumberOrArray
            ) -> None:
                """add an amount to a field at an interpolated position

                Args:
                    data (:class:`~numpy.ndarray`):
                        The values at the grid points
                    point (:class:`~numpy.ndarray`):
                        Coordinates of a single point in the grid coordinate system
                    amount (Number or :class:`~numpy.ndarray`):
                        The amount that will be added to the data. This value describes
                        an integrated quantity (given by the field value times the
                        discretization volume). This is important for consistency with
                        different discretizations and in particular grids with
                        non-uniform discretizations
                """
                # determine surrounding points and their weights
                c_xli, c_xhi, w_xl, w_xh = data_x(point[0])
                c_yli, c_yhi, w_yl, w_yh = data_y(point[1])
                c_zli, c_zhi, w_zl, w_zh = data_z(point[2])

                if c_xli == -42 or c_yli == -42 or c_zli == -42:  # out of bounds
                    raise DomainError("Point lies outside the grid domain")

                cell_vol = cell_volume(c_xli, c_yli, c_zli)
                data[..., c_xli, c_yli, c_zli] += w_xl * w_yl * w_zl * amount / cell_vol
                cell_vol = cell_volume(c_xli, c_yli, c_zhi)
                data[..., c_xli, c_yli, c_zhi] += w_xl * w_yl * w_zh * amount / cell_vol

                cell_vol = cell_volume(c_xli, c_yhi, c_zli)
                data[..., c_xli, c_yhi, c_zli] += w_xl * w_yh * w_zl * amount / cell_vol
                cell_vol = cell_volume(c_xli, c_yhi, c_zhi)
                data[..., c_xli, c_yhi, c_zhi] += w_xl * w_yh * w_zh * amount / cell_vol

                cell_vol = cell_volume(c_xhi, c_yli, c_zli)
                data[..., c_xhi, c_yli, c_zli] += w_xh * w_yl * w_zl * amount / cell_vol
                cell_vol = cell_volume(c_xhi, c_yli, c_zhi)
                data[..., c_xhi, c_yli, c_zhi] += w_xh * w_yl * w_zh * amount / cell_vol

                cell_vol = cell_volume(c_xhi, c_yhi, c_zli)
                data[..., c_xhi, c_yhi, c_zli] += w_xh * w_yh * w_zl * amount / cell_vol
                cell_vol = cell_volume(c_xhi, c_yhi, c_zhi)
                data[..., c_xhi, c_yhi, c_zhi] += w_xh * w_yh * w_zh * amount / cell_vol

        else:
            raise NotImplementedError(
                f"Compiled interpolation not implemented for dimension {self.num_axes}"
            )

        return insert  # type: ignore

    def make_integrator(self) -> Callable[[np.ndarray], NumberOrArray]:
        """return function that can be used to integrates discretized data over the grid

        If this function is used in a multiprocessing run (using MPI), the integrals are
        performed on all subgrids and then accumulated. Each process then receives the
        same value representing the global integral.

        Returns:
            callable: A function that takes a numpy array and returns the integral with
            the correct weights given by the cell volumes.
        """
        num_axes = self.num_axes
        # cell volume varies with position
        get_cell_volume = self.make_cell_volume_compiled(flat_index=True)

        def integrate_local(arr: np.ndarray) -> NumberOrArray:
            """integrates data over a grid using numpy"""
            amounts = arr * self.cell_volumes
            return amounts.sum(axis=tuple(range(-num_axes, 0, 1)))  # type: ignore

        @overload(integrate_local)
        def ol_integrate_local(
            arr: np.ndarray,
        ) -> Callable[[np.ndarray], NumberOrArray]:
            """integrates data over a grid using numba"""
            if arr.ndim == num_axes:
                # `arr` is a scalar field
                grid_shape = self.shape

                def impl(arr: np.ndarray) -> Number:
                    """integrate a scalar field"""
                    assert arr.shape == grid_shape
                    total = 0
                    for i in range(arr.size):
                        total += get_cell_volume(i) * arr.flat[i]
                    return total

            else:
                # `arr` is a tensorial field with rank >= 1
                tensor_shape = (self.dim,) * (arr.ndim - num_axes)
                data_shape = tensor_shape + self.shape

                def impl(arr: np.ndarray) -> np.ndarray:  # type: ignore
                    """integrate a tensorial field"""
                    assert arr.shape == data_shape
                    total = np.zeros(tensor_shape)
                    for idx in np.ndindex(*tensor_shape):
                        arr_comp = arr[idx]
                        for i in range(arr_comp.size):
                            total[idx] += get_cell_volume(i) * arr_comp.flat[i]
                    return total

            return impl

        # deal with MPI multiprocessing
        if self._mesh is None or len(self._mesh) == 1:
            # standard case of a single integral
            @jit
            def integrate_global(arr: np.ndarray) -> NumberOrArray:
                """integrate data

                Args:
                    arr (:class:`~numpy.ndarray`): discretized data on grid
                """
                return integrate_local(arr)

        else:
            # we are in a parallel run, so we need to gather the sub-integrals from all
            # subgrids in the grid mesh
            from ..tools.mpi import mpi_allreduce

            @jit
            def integrate_global(arr: np.ndarray) -> NumberOrArray:
                """integrate data over MPI parallelized grid

                Args:
                    arr (:class:`~numpy.ndarray`): discretized data on grid
                """
                integral = integrate_local(arr)
                return mpi_allreduce(integral)  # type: ignore

        return integrate_global  # type: ignore


def registered_operators() -> Dict[str, List[str]]:
    """returns all operators that are currently defined

    Returns:
        dict: a dictionary with the names of the operators defined for each grid class
    """
    return {
        name: sorted(cls.operators)
        for name, cls in GridBase._subclasses.items()
        if not (name.endswith("Base") or hasattr(cls, "deprecated") and cls.deprecated)
    }



r"""
Heterogeneous boundary conditions
=================================

This example implements a `spatially coupled SIR model 
<https://en.wikipedia.org/wiki/Compartmental_models_in_epidemiology>`_ with 
the following dynamics for the density of susceptible, infected, and recovered
individuals:

.. math::

    \partial_t s &= D \nabla^2 s - \beta is \\
    \partial_t i &= D \nabla^2 i + \beta is - \gamma i \\
    \partial_t r &= D \nabla^2 r + \gamma i

Here, :math:`D` is the diffusivity, :math:`\beta` the infection rate, and
:math:`\gamma` the recovery rate.
"""

import numpy as np

from pde import CartesianGrid, DiffusionPDE, ScalarField

# define grid and an initial state
grid = CartesianGrid([[-5, 5], [-5, 5]], 32)
field = ScalarField(grid)


# define the boundary conditions, which here are calculated from a function
def bc_value(adjacent_value, dx, x, y, t):
    """return boundary value"""
    return np.sign(x)


bc_x = "derivative"
bc_y = ["derivative", {"value_expression": bc_value}]

# define and solve a simple diffusion equation
eq = DiffusionPDE(bc=[bc_x, bc_y])
res = eq.solve(field, t_range=10, dt=0.01, backend="numpy")
res.plot()



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np

from pde.tools.numba import Counter, flat_idx, jit, numba_environment


def test_environment():
    """test function signature checks"""
    assert isinstance(numba_environment(), dict)


def test_flat_idx():
    """test flat_idx function"""
    # testing the numpy version
    assert flat_idx(2, 1) == 2
    assert flat_idx(np.arange(2), 1) == 1
    assert flat_idx(np.arange(4).reshape(2, 2), 1) == 1

    # testing the numba compiled version
    @jit
    def get_data(data):
        return flat_idx(data, 1)

    assert get_data(2) == 2
    assert get_data(np.arange(2)) == 1
    assert get_data(np.arange(4).reshape(2, 2)) == 1


def test_counter():
    """test Counter implementation"""
    c1 = Counter()
    assert int(c1) is 0
    assert c1 == 0
    assert str(c1) == "0"

    c1.increment()
    assert int(c1) is 1

    c1 += 2
    assert int(c1) is 3

    c2 = Counter(3)
    assert c1 is not c2
    assert c1 == c2



"""
This file is used to configure the test environment when running py.test

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import matplotlib.pyplot as plt
import numpy as np
import pytest

from pde.tools.misc import module_available
from pde.tools.numba import random_seed


@pytest.fixture(scope="function", autouse=True)
def setup_and_teardown():
    """helper function adjusting environment before and after tests"""
    # raise all underflow errors
    np.seterr(all="raise", under="ignore")

    # run the actual test
    yield

    # clean up open matplotlib figures after the test
    plt.close("all")


@pytest.fixture(scope="function", autouse=False, name="rng")
def init_random_number_generators():
    """get a random number generator and set the seed of the random number generator

    The function returns an instance of :func:`~numpy.random.default_rng()` and
    initializes the default generators of both :mod:`numpy` and :mod:`numba`.
    """
    random_seed()
    return np.random.default_rng(0)


def pytest_configure(config):
    """add markers to the configuration"""
    config.addinivalue_line("markers", "interactive: test is interactive")
    config.addinivalue_line("markers", "multiprocessing: test requires multiprocessing")
    config.addinivalue_line("markers", "slow: test runs slowly")


def pytest_addoption(parser):
    """pytest hook to add command line options parsed by pytest"""
    parser.addoption(
        "--runslow",
        action="store_true",
        default=False,
        help="also run tests marked by `slow`",
    )
    parser.addoption(
        "--runinteractive",
        action="store_true",
        default=False,
        help="also run tests marked by `interactive`",
    )
    parser.addoption(
        "--use_mpi",
        action="store_true",
        default=False,
        help="only run tests marked by `multiprocessing`",
    )


def pytest_collection_modifyitems(config, items):
    """pytest hook to filter a collection of tests"""
    # parse options provided to py.test
    running_cov = config.getvalue("--cov")
    runslow = config.getoption("--runslow", default=False)
    runinteractive = config.getoption("--runinteractive", default=False)
    use_mpi = config.getoption("--use_mpi", default=False)
    has_numba_mpi = module_available("numba_mpi") and module_available("mpi4py")

    # prepare markers
    skip_cov = pytest.mark.skip(reason="skipped during coverage run")
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    skip_interactive = pytest.mark.skip(reason="need --runinteractive option to run")
    skip_serial = pytest.mark.skip(reason="serial test, but --use_mpi option was set")
    skip_mpi = pytest.mark.skip(reason="mpi test, but `numba_mpi` not available")

    # check each test item
    for item in items:
        if "no_cover" in item.keywords and running_cov:
            item.add_marker(skip_cov)
        if "slow" in item.keywords and not runslow:
            item.add_marker(skip_slow)
        if "interactive" in item.keywords and not runinteractive:
            item.add_marker(skip_interactive)

        if "multiprocessing" in item.keywords and not has_numba_mpi:
            item.add_marker(skip_mpi)
        if use_mpi and "multiprocessing" not in item.keywords:
            item.add_marker(skip_serial)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import CartesianGrid, CylindricalSymGrid, ScalarField
from pde.grids.boundaries.local import NeumannBC


@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("r_inner", [0, 2])
def test_cylindrical_grid(periodic, r_inner, rng):
    """test simple cylindrical grid"""
    grid = CylindricalSymGrid((r_inner, 4), (-1, 2), (8, 9), periodic_z=periodic)
    if r_inner == 0:
        assert grid == CylindricalSymGrid(4, (-1, 2), (8, 9), periodic_z=periodic)
    rs, zs = grid.axes_coords

    assert grid.dim == 3
    assert grid.numba_type == "f8[:, :]"
    assert grid.shape == (8, 9)
    assert grid.length == pytest.approx(3)
    assert grid.discretization[1] == pytest.approx(1 / 3)
    assert grid.volume == pytest.approx(3 * np.pi * (4**2 - r_inner**2))
    assert not grid.uniform_cell_volumes
    assert grid.volume == pytest.approx(grid.integrate(1))
    np.testing.assert_allclose(zs, np.linspace(-1 + 1 / 6, 2 - 1 / 6, 9))

    if r_inner == 0:
        assert grid.discretization[0] == pytest.approx(0.5)
        np.testing.assert_array_equal(grid.discretization, np.array([0.5, 1 / 3]))
        np.testing.assert_allclose(rs, np.linspace(0.25, 3.75, 8))
    else:
        assert grid.discretization[0] == pytest.approx(0.25)
        np.testing.assert_array_equal(grid.discretization, np.array([0.25, 1 / 3]))
        np.testing.assert_allclose(rs, np.linspace(2.125, 3.875, 8))

    assert grid.contains_point(grid.get_random_point(coords="cartesian", rng=rng))
    ps = [grid.get_random_point(coords="cartesian", rng=rng) for _ in range(2)]
    assert all(grid.contains_point(ps))
    ps = grid.get_random_point(coords="cartesian", boundary_distance=1.49, rng=rng)
    assert grid.contains_point(ps)
    assert "laplace" in grid.operators


def test_cylindrical_to_cartesian():
    """test conversion of cylindrical grid to Cartesian"""
    expr_cyl = "cos(z / 2) / (1 + r**2)"
    expr_cart = expr_cyl.replace("r**2", "(x**2 + y**2)")

    z_range = (-np.pi, 2 * np.pi)
    grid_cyl = CylindricalSymGrid(10, z_range, (16, 33))
    pf_cyl = ScalarField.from_expression(grid_cyl, expression=expr_cyl)

    grid_cart = CartesianGrid([[-7, 7], [-6, 7], z_range], [16, 16, 16])
    pf_cart1 = pf_cyl.interpolate_to_grid(grid_cart)
    pf_cart2 = ScalarField.from_expression(grid_cart, expression=expr_cart)
    np.testing.assert_allclose(pf_cart1.data, pf_cart2.data, atol=0.1)


def test_setting_boundary_conditions():
    """test various versions of settings bcs for cylindrical grids"""
    grid = CylindricalSymGrid(1, [0, 1], [2, 2], periodic_z=False)
    grid.get_boundary_conditions("auto_periodic_neumann")
    grid.get_boundary_conditions(["derivative", "derivative"])
    with pytest.raises(ValueError):
        grid.get_boundary_conditions(["derivative"])
    with pytest.raises(ValueError):
        grid.get_boundary_conditions(["derivative"] * 3)
    with pytest.raises(RuntimeError):
        grid.get_boundary_conditions(["derivative", "periodic"])

    b_inner = NeumannBC(grid, 0, upper=False)
    assert grid.get_boundary_conditions("auto_periodic_neumann")[0].low == b_inner
    assert grid.get_boundary_conditions({"value": 2})[0].low != b_inner

    grid = CylindricalSymGrid(1, [0, 1], [2, 2], periodic_z=True)
    grid.get_boundary_conditions("auto_periodic_neumann")
    grid.get_boundary_conditions(["derivative", "periodic"])
    with pytest.raises(RuntimeError):
        grid.get_boundary_conditions(["derivative", "derivative"])


@pytest.mark.parametrize("periodic", [True, False])
def test_polar_conversion(periodic):
    """test conversion to polar coordinates"""
    grid = CylindricalSymGrid(1, [-1, 1], [5, 5], periodic_z=periodic)
    dists = grid.polar_coordinates_real([0, 0, 0])
    assert np.all(0.09 <= dists)
    assert np.any(dists < 0.11)
    assert np.all(dists <= np.sqrt(2))
    assert np.any(dists > 0.8 * np.sqrt(2))



"""
Defines a vectorial field over a grid

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
)

import numba as nb
import numpy as np
from numba.extending import overload, register_jitable
from numpy.typing import DTypeLike

from ..grids.base import DimensionError, GridBase
from ..tools.docstrings import fill_in_docstring
from ..tools.misc import get_common_dtype
from ..tools.numba import get_common_numba_dtype, jit
from ..tools.typing import NumberOrArray
from .base import DataFieldBase
from .scalar import ScalarField

if TYPE_CHECKING:
    from ..grids.boundaries.axes import BoundariesData  # @UnusedImport
    from .tensorial import Tensor2Field  # @UnusedImport


class VectorField(DataFieldBase):
    """Vector field discretized on a grid"""

    rank = 1

    @classmethod
    def from_scalars(
        cls,
        fields: List[ScalarField],
        *,
        label: Optional[str] = None,
        dtype: Optional[DTypeLike] = None,
    ) -> VectorField:
        """create a vector field from a list of ScalarFields

        Note that the data of the scalar fields is copied in the process

        Args:
            fields (list):
                The list of (compatible) scalar fields
            label (str, optional):
                Name of the returned field
            dtype (numpy dtype):
                The data type of the field. If omitted, it will be determined from
                `data` automatically.

        Returns:
            :class:`VectorField`: the resulting vector field
        """
        grid = fields[0].grid

        if grid.dim != len(fields):
            raise DimensionError(
                f"Grid dimension and number of scalar fields differ ({grid.dim} != "
                f"{len(fields)})"
            )

        data = []
        for field in fields:
            assert field.grid.compatible_with(grid)
            data.append(field.data)

        return cls(grid, data, label=label, dtype=dtype)

    @classmethod
    @fill_in_docstring
    def from_expression(
        cls,
        grid: GridBase,
        expressions: Sequence[str],
        *,
        user_funcs: Optional[Dict[str, Callable]] = None,
        consts: Optional[Dict[str, NumberOrArray]] = None,
        label: Optional[str] = None,
        dtype: Optional[DTypeLike] = None,
    ) -> VectorField:
        """create a vector field on a grid from given expressions

        Warning:
            {WARNING_EXEC}

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                Grid defining the space on which this field is defined
            expressions (list of str):
                A list of mathematical expression, one for each component of the vector
                field. The expressions determine the values as a function of the
                position on the grid. The expressions may contain standard mathematical
                functions and they may depend on the axes labels of the grid.
                More information can be found in the
                :ref:`expression documentation <documentation-expressions>`.
            user_funcs (dict, optional):
                A dictionary with user defined functions that can be used in the
                expression
            consts (dict, optional):
                A dictionary with user defined constants that can be used in the
                expression. The values of these constants should either be numbers or
                :class:`~numpy.ndarray`.
            label (str, optional):
                Name of the field
            dtype (numpy dtype):
                The data type of the field. If omitted, it will be determined from
                `data` automatically.
        """
        from ..tools.expressions import ScalarExpression

        if isinstance(expressions, str) or len(expressions) != grid.dim:
            axes_names = grid.axes + grid.axes_symmetric
            raise DimensionError(
                f"Expected {grid.dim} expressions for the coordinates {axes_names}."
            )

        # obtain the coordinates of the grid points
        points = [grid.cell_coords[..., i] for i in range(grid.num_axes)]

        # evaluate all vector components at all points
        data = []
        for expression in expressions:
            expr = ScalarExpression(
                expression=expression,
                signature=grid.axes,
                user_funcs=user_funcs,
                consts=consts,
            )
            values = np.broadcast_to(expr(*points), grid.shape)
            data.append(values)

        # create vector field from the data
        return cls(grid=grid, data=data, label=label, dtype=dtype)

    def __getitem__(self, key: Union[int, str]) -> ScalarField:
        """extract a component of the VectorField"""
        return ScalarField(
            self.grid,
            data=self._data_full[self.grid.get_axis_index(key)],
            with_ghost_cells=True,
        )

    def __setitem__(
        self, key: Union[int, str], value: Union[NumberOrArray, ScalarField]
    ):
        """set a component of the VectorField"""
        idx = self.grid.get_axis_index(key)
        if isinstance(value, ScalarField):
            self.grid.assert_grid_compatible(value.grid)
            self.data[idx] = value.data
        else:
            self.data[idx] = value

    def dot(
        self,
        other: Union[VectorField, "Tensor2Field"],
        out: Union[ScalarField, VectorField, None] = None,
        *,
        conjugate: bool = True,
        label: str = "dot product",
    ) -> Union[ScalarField, VectorField]:
        """calculate the dot product involving a vector field

        This supports the dot product between two vectors fields as well as the
        product between a vector and a tensor. The resulting fields will be a
        scalar or vector, respectively.

        Args:
            other (VectorField or Tensor2Field):
                the second field
            out (ScalarField or VectorField, optional):
                Optional field to which the result is written.
            conjugate (bool):
                Whether to use the complex conjugate for the second operand
            label (str, optional):
                Name of the returned field

        Returns:
            :class:`~pde.fields.scalar.ScalarField` or
            :class:`~pde.fields.vectorial.VectorField`: result of applying the operator
        """
        from .tensorial import Tensor2Field  # @Reimport

        # check input
        self.grid.assert_grid_compatible(other.grid)
        if isinstance(other, VectorField):
            result_type = ScalarField
        elif isinstance(other, Tensor2Field):
            result_type = VectorField  # type: ignore
        else:
            raise TypeError("Second term must be a vector or tensor field")

        if out is None:
            out = result_type(self.grid, dtype=get_common_dtype(self, other))
        else:
            assert isinstance(out, result_type), f"`out` must be {result_type}"
            self.grid.assert_grid_compatible(out.grid)

        # calculate the result
        other_data = other.data.conjugate() if conjugate else other.data
        np.einsum("i...,i...->...", self.data, other_data, out=out.data)
        if label is not None:
            out.label = label

        return out

    __matmul__ = dot  # support python @-syntax for matrix multiplication

    def outer_product(
        self,
        other: VectorField,
        out: Optional["Tensor2Field"] = None,
        *,
        label: Optional[str] = None,
    ) -> "Tensor2Field":
        """calculate the outer product of this vector field with another

        Args:
            other (:class:`~pde.fields.vectorial.VectorField`):
                The second vector field
            out (:class:`~pde.fields.tensorial.Tensor2Field`, optional):
                Optional tensorial field to which the  result is written.
            label (str, optional):
                Name of the returned field

        Returns:
            :class:`~pde.fields.tensorial.Tensor2Field`: result of the operation
        """
        from .tensorial import Tensor2Field  # @Reimport

        self.assert_field_compatible(other)

        if out is None:
            out = Tensor2Field(self.grid)
        else:
            self.grid.assert_grid_compatible(out.grid)

        # calculate the result
        np.einsum("i...,j...->ij...", self.data, other.data, out=out.data)
        if label is not None:
            out.label = label

        return out

    def make_outer_prod_operator(
        self, backend: Literal["numpy", "numba"] = "numba"
    ) -> Callable[[np.ndarray, np.ndarray, Optional[np.ndarray]], np.ndarray]:
        """return operator calculating the outer product of two vector fields

        Warning:
            This function does not check types or dimensions.

        Args:
            backend (str):
                The backend (e.g., 'numba' or 'scipy') used for this operator.

        Returns:
            function that takes two instance of :class:`~numpy.ndarray`, which contain
            the discretized data of the two operands. An optional third argument can
            specify the output array to which the result is written. Note that the
            returned function is jitted with numba for speed.
        """

        def outer(
            a: np.ndarray, b: np.ndarray, out: Optional[np.ndarray] = None
        ) -> np.ndarray:
            """calculate the outer product using numpy"""
            return np.einsum("i...,j...->ij...", a, b, out=out)

        if backend == "numpy":
            # return the bare dot operator without the numba-overloaded version
            return outer

        elif backend == "numba":
            # overload `outer` with a numba-compiled version

            dim = self.grid.dim
            num_axes = self.grid.num_axes

            def check_rank(arr: Union[nb.types.Type, nb.types.Optional]) -> None:
                """determine rank of field with type `arr`"""
                arr_typ = arr.type if isinstance(arr, nb.types.Optional) else arr
                if not isinstance(arr_typ, (np.ndarray, nb.types.Array)):
                    raise nb.errors.TypingError(
                        f"Arguments must be array, not  {arr_typ.__class__}"
                    )
                assert arr_typ.ndim == 1 + num_axes

            # create the inner function calculating the outer product
            @register_jitable
            def calc(a: np.ndarray, b: np.ndarray, out: np.ndarray) -> np.ndarray:
                """calculate outer product between fields `a` and `b`"""
                for i in range(0, dim):
                    for j in range(0, dim):
                        out[i, j, :] = a[i] * b[j]
                return out

            @overload(outer, inline="always")
            def outer_ol(
                a: np.ndarray, b: np.ndarray, out: Optional[np.ndarray] = None
            ) -> np.ndarray:
                """numba implementation to calculate outer product between two fields"""
                # get (and check) rank of the input arrays
                check_rank(a)
                check_rank(b)
                in_shape = (dim,) + self.grid.shape
                out_shape = (dim, dim) + self.grid.shape

                if isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                    # function is called without `out` -> allocate memory
                    dtype = get_common_numba_dtype(a, b)

                    def outer_impl(
                        a: np.ndarray, b: np.ndarray, out: Optional[np.ndarray] = None
                    ) -> np.ndarray:
                        """helper function allocating output array"""
                        assert a.shape == b.shape == in_shape
                        out = np.empty(out_shape, dtype=dtype)
                        calc(a, b, out)
                        return out

                else:
                    # function is called with `out` argument -> reuse `out` array

                    def outer_impl(
                        a: np.ndarray, b: np.ndarray, out: Optional[np.ndarray] = None
                    ) -> np.ndarray:
                        """helper function without allocating output array"""
                        # check input
                        assert a.shape == b.shape == in_shape
                        assert out.shape == out_shape  # type: ignore
                        calc(a, b, out)
                        return out  # type: ignore

                return outer_impl  # type: ignore

            @jit
            def outer_compiled(
                a: np.ndarray, b: np.ndarray, out: Optional[np.ndarray] = None
            ) -> np.ndarray:
                """numba implementation to calculate outer product between two fields"""
                return outer(a, b, out)

            return outer_compiled  # type: ignore

        else:
            raise ValueError(f"Unsupported backend `{backend}")

    @fill_in_docstring
    def divergence(
        self, bc: Optional[BoundariesData], out: Optional[ScalarField] = None, **kwargs
    ) -> ScalarField:
        """apply divergence operator and return result as a field

        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES_OPTIONAL}
            out (ScalarField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field
            **kwargs:
                Additional arguments affecting how the operator behaves.

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: Divergence of the field
        """
        return self.apply_operator("divergence", bc=bc, out=out, **kwargs)  # type: ignore

    @fill_in_docstring
    def gradient(
        self,
        bc: Optional[BoundariesData],
        out: Optional["Tensor2Field"] = None,
        **kwargs,
    ) -> "Tensor2Field":
        r"""apply vector gradient operator and return result as a field

        The vector gradient field is a tensor field :math:`t_{\alpha\beta}` that
        specifies the derivatives of the vector field :math:`v_\alpha` with respect to
        all coordinates :math:`x_\beta`.

        Args:
            bc:
                The boundary conditions applied to the field. Boundary conditions need
                to determine all components of the vector field.
                {ARG_BOUNDARIES_OPTIONAL}
            out (VectorField, optional):
                Optional vector field to which the result is written.
            label (str, optional):
                Name of the returned field
            **kwargs:
                Additional arguments affecting how the operator behaves.

        Returns:
            :class:`~pde.fields.tensorial.Tensor2Field`: Gradient of the field
        """
        return self.apply_operator("vector_gradient", bc=bc, out=out, **kwargs)  # type: ignore

    @fill_in_docstring
    def laplace(
        self, bc: Optional[BoundariesData], out: Optional[VectorField] = None, **kwargs
    ) -> VectorField:
        r"""apply vector Laplace operator and return result as a field

        The vector Laplacian is a vector field :math:`L_\alpha` containing the second
        derivatives of the vector field :math:`v_\alpha` with respect to the coordinates
        :math:`x_\beta`:

        .. math::
            L_\alpha = \sum_\beta
                \frac{\partial^2 v_\alpha}{\partial x_\beta \partial x_\beta}

        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES_OPTIONAL}
            out (VectorField, optional):
                Optional vector field to which the  result is written.
            label (str, optional):
                Name of the returned field
            **kwargs:
                Additional arguments affecting how the operator behaves.

        Returns:
            :class:`~pde.fields.vectorial.VectorField`: Laplacian of the field
        """
        return self.apply_operator("vector_laplace", bc=bc, out=out, **kwargs)  # type: ignore

    @property
    def integral(self) -> np.ndarray:
        """:class:`~numpy.ndarray`: integral of each component over space"""
        return self.grid.integrate(self.data)  # type: ignore

    def to_scalar(
        self,
        scalar: Union[str, int] = "auto",
        *,
        label: Optional[str] = "scalar `{scalar}`",
    ) -> ScalarField:
        """return a scalar field by applying `method`

        Args:
            scalar (str):
                Choose the method to use. Possible  choices are `norm`, `max`, `min`,
                `squared_sum`, `norm_squared`, or an integer specifying which component
                is returned (indexing starts at `0`). The default value `auto` picks the
                method automatically: The first (and only) component is returned for
                real fields on one-dimensional spaces, while the norm of the vector is
                returned otherwise.
            label (str, optional):
                Name of the returned field

        Returns:
            :class:`pde.fields.scalar.ScalarField`:
                The scalar field after applying the operation
        """
        if scalar == "auto":
            if self.grid.dim > 1 or np.iscomplexobj(self.data):
                scalar = "norm"
            else:
                scalar = 0  # return the field unchanged

        if isinstance(scalar, int):
            data = self.data[scalar]

        elif scalar == "norm":
            data = np.linalg.norm(self.data, axis=0)

        elif scalar == "max":
            data = np.max(self.data, axis=0)

        elif scalar == "min":
            data = np.min(self.data, axis=0)

        elif scalar == "squared_sum":
            data = np.sum(self.data**2, axis=0)

        elif scalar == "norm_squared":
            data = np.sum(self.data * self.data.conjugate(), axis=0)

        else:
            raise ValueError(f"Unknown method `{scalar}` for `to_scalar`")

        if label is not None:
            label = label.format(scalar=scalar)

        return ScalarField(self.grid, data, label=label)

    def get_vector_data(
        self, transpose: bool = False, max_points: Optional[int] = None, **kwargs
    ) -> Dict[str, Any]:
        r"""return data for a vector plot of the field

        Args:
            transpose (bool):
                Determines whether the transpose of the data should be plotted.
            max_points (int):
                The maximal number of points that is used along each axis. This
                option can be used to sub-sample the data.
            \**kwargs:
                Additional parameters forwarded to `grid.get_image_data`

        Returns:
            dict: Information useful for plotting an vector field
        """
        # TODO: Handle Spherical and Cartesian grids, too. This could be
        # implemented by adding a get_vector_data method to the grids
        if self.grid.dim == 2:
            vx = self[0].get_image_data(**kwargs)
            vy = self[1].get_image_data(**kwargs)
            data = vx  # use one of the fields to extract basic information
            data["data_x"] = vx.pop("data")
            data["data_y"] = vy["data"]
            data["title"] = self.label

        else:
            raise NotImplementedError("Only supports 2d grids")

        # transpose the data if requested
        if transpose:
            data["x"], data["y"] = data["y"], data["x"]
            data["data_x"], data["data_y"] = data["data_y"].T, data["data_x"].T
            data["label_x"], data["label_y"] = data["label_y"], data["label_x"]
            data["extent"] = data["extent"][2:] + data["extent"][:2]

        # reduce the sampling of the vector points
        if max_points is not None:
            shape = data["data_x"].shape
            for axis, size in enumerate(shape):
                if size > max_points:
                    # sub-sample the data
                    idx_f = np.linspace(0, size - 1, max_points)
                    idx_i = np.round(idx_f).astype(int)

                    data["data_x"] = np.take(data["data_x"], idx_i, axis=axis)
                    data["data_y"] = np.take(data["data_y"], idx_i, axis=axis)
                    if axis == 0:
                        data["x"] = data["x"][idx_i]
                    elif axis == 1:
                        data["y"] = data["y"][idx_i]
                    else:
                        raise RuntimeError("Only supports 2d grids")

        data["shape"] = data["data_x"].shape
        data["size"] = data["data_x"].size

        return data

    def _get_napari_layer_data(  # type: ignore
        self, max_points: Optional[int] = None, args: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """returns data for plotting on a single napari layer

        Args:
            max_points (int):
                The maximal number of points that is used along each axis. This option
                can be used to subsample the data.
            args (dict):
                Additional arguments returned in the result, which affect how the layer
                is shown.

        Returns:
            dict: all the information necessary to plot this field
        """
        result = {} if args is None else args.copy()

        # extract the vector components in the format required by napari
        data = self.get_vector_data(max_points=max_points)
        vectors = np.empty((data["size"], 2, 2))
        xs, ys = np.meshgrid(data["x"], data["y"], indexing="ij")
        vectors[:, 0, 0] = xs.flat
        vectors[:, 0, 1] = ys.flat
        vectors[:, 1, 0] = data["data_x"].flat
        vectors[:, 1, 1] = data["data_y"].flat

        result["type"] = "vectors"
        result["data"] = vectors
        return result



"""
Provides support for mypy type checking of the package

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import TYPE_CHECKING, Literal, Protocol, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike  # @UnusedImport

if TYPE_CHECKING:
    from ..grids.base import GridBase  # @UnusedImport

Real = Union[int, float]
Number = Union[Real, complex]
NumberOrArray = Union[Number, np.ndarray]
FloatNumerical = Union[float, np.ndarray]
BackendType = Literal["auto", "numpy", "numba"]


class OperatorType(Protocol):
    """an operator that acts on an array"""

    def __call__(self, arr: np.ndarray, out: np.ndarray) -> None:
        ...


class OperatorFactory(Protocol):
    """a factory function that creates an operator for a particular grid"""

    def __call__(self, grid: "GridBase", **kwargs) -> OperatorType:
        ...


class AdjacentEvaluator(Protocol):
    def __call__(
        self, arr_1d: np.ndarray, i_point: int, bc_idx: Tuple[int, ...]
    ) -> float:
        ...


class CellVolume(Protocol):
    def __call__(self, *args: int) -> float:
        ...


class GhostCellSetter(Protocol):
    def __call__(self, data_full: np.ndarray, args=None) -> None:
        ...


class VirtualPointEvaluator(Protocol):
    def __call__(self, arr: np.ndarray, idx: Tuple[int, ...], args=None) -> float:
        ...



"""
Simple sphinx plug-in that simplifies  type information in function signatures
"""

import re

# simple (literal) replacement rules
REPLACEMENTS = [
    # numbers and numerical arrays
    (
        "Union[int, float, complex, numpy.generic, numpy.ndarray, "
        "Sequence[Union[int, float, complex, numpy.generic, numpy.ndarray]], "
        "Sequence[Sequence[Any]]]",
        "ArrayLike",
    ),
    (
        "Union[int, float, complex, numpy.ndarray, "
        "Sequence[Union[int, float, complex, numpy.ndarray]], "
        "Sequence[Sequence[Any]]]",
        "ArrayLike",
    ),
    ("Union[int, float, complex, numpy.generic, numpy.ndarray]", "NumberOrArray"),
    ("Union[int, float, complex, numpy.ndarray]", "NumberOrArray"),
    ("Union[int, float, complex]", "Number"),
    (
        "Optional[Union[_SupportsArray[dtype], _NestedSequence[_SupportsArray[dtype]], "
        "bool, int, float, complex, str, bytes, _NestedSequence[Union[bool, int, "
        "float, complex, str, bytes]]]]",
        "NumberOrArray",
    ),
    (
        "Union[dtype[Any], None, Type[Any], _SupportsDType[dtype[Any]], str, "
        "Tuple[Any, int], Tuple[Any, Union[SupportsIndex, Sequence[SupportsIndex]]], "
        "List[Any], _DTypeDict, Tuple[Any, Any]]",
        "DType",
    ),
    # Remove some unnecessary information in favor of a more compact style
    ("Dict[KT, VT]", "dict"),
    ("Dict[str, Any]", "dict"),
    ("Optional[str]", "str"),
    ("Optional[int]", "int"),
    ("Optional[float]", "float"),
    ("Optional[numpy.ndarray]", "numpy.ndarray"),
    ("Optional[dict]", "dict"),
    ("Optional[Dict[str, Any]]", "dict"),
    # Complex types describing the boundary conditions
    ("Union[dict, str, BCBase]", "BoundaryData"),
    (
        "Union[Dict[str, BoundaryData], "
        "BoundaryData, Tuple[BoundaryData, BoundaryData]]",
        "BoundaryPairData",
    ),
    (
        "Union[Dict[str, BoundaryData], dict, str, BCBase, "
        "Tuple[BoundaryData, BoundaryData]]",
        "BoundaryPairData",
    ),
    ("Union[BoundaryPairData, Sequence[BoundaryPairData]]", "BoundariesData"),
    (
        "Union[Dict[str, BoundaryData], dict, str, BCBase, "
        "Tuple[BoundaryData, BoundaryData], Sequence[BoundaryPairData]]",
        "BoundariesData",
    ),
    (
        "Union[dict, str, BCBase, Tuple[Union[dict, str, BCBase], "
        "Union[dict, str, BCBase]], Sequence[Union[dict, str, BCBase, "
        "Tuple[Union[dict, str, BCBase], Union[dict, str, BCBase]]]]]",
        "BoundaryConditionData",
    ),
    (
        "Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
        "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]], "
        "Sequence[Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
        "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]]]]]",
        "BoundaryConditionData",
    ),
    (
        "Union[dict, str, BCBase, Tuple[Union[dict, str, BCBase], "
        "Union[dict, str, BCBase]], Sequence[Union[dict, str, BCBase, "
        "Tuple[Union[dict, str, BCBase], Union[dict, str, BCBase]]]], "
        "Sequence[BoundaryConditionData]]",
        "BoundaryConditionData",
    ),
    (
        "Union[Dict[str, BoundaryData], dict, str, BCBase, "
        "Tuple[BoundaryData, BoundaryData], "
        "Sequence[BoundaryPairData], Sequence[BoundariesData]]",
        "BoundariesDataList",
    ),
    (
        "Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
        "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]], "
        "Sequence[Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
        "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]]]], "
        "Sequence[BoundaryConditionData]]",
        "BoundariesDataList",
    ),
    # Other compound data types
    ("Union[List[Union[TrackerBase, str]], TrackerBase, str, None]", "TrackerData"),
    (
        "Optional[Union[List[Union[TrackerBase, str]], TrackerBase, str]]",
        "TrackerData",
    ),
]


# replacement rules based on regular expressions
REPLACEMENTS_REGEX = {
    # remove full package path and only leave the module/class identifier
    "pde\.(\w+\.)*": "",
    "typing\.": "",
}


def process_signature(
    app, what: str, name: str, obj, options, signature, return_annotation
):
    """Process signature by applying replacement rules"""

    def process(sig_obj):
        """process the signature object"""
        if sig_obj is not None:
            for key, value in REPLACEMENTS_REGEX.items():
                sig_obj = re.sub(key, value, sig_obj)
            for key, value in REPLACEMENTS:
                sig_obj = sig_obj.replace(key, value)
        return sig_obj

    signature = process(signature)
    return_annotation = process(return_annotation)

    return signature, return_annotation


def setup(app):
    """set up hooks for this sphinx plugin"""
    app.connect("autodoc-process-signature", process_signature)



"""
Stochastic simulation
=====================

This example illustrates how a stochastic simulation can be done.
"""

from pde import KPZInterfacePDE, MemoryStorage, ScalarField, UnitGrid, plot_kymograph

grid = UnitGrid([64])  # generate grid
state = ScalarField.random_harmonic(grid)  # generate initial condition

eq = KPZInterfacePDE(noise=1)  # define the SDE
storage = MemoryStorage()
eq.solve(state, t_range=10, dt=0.01, tracker=storage.tracker(0.5))
plot_kymograph(storage)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import os
import pickle

import numpy as np
import pytest

from pde import Controller, ExplicitSolver, MemoryStorage, ScalarField, UnitGrid
from pde.pdes import AllenCahnPDE, CahnHilliardPDE, DiffusionPDE
from pde.tools.misc import module_available
from pde.trackers import get_named_trackers, trackers
from pde.trackers.base import TrackerBase
from pde.trackers.interrupts import ConstantInterrupts
from pde.visualization.movies import Movie


def test_plot_tracker(tmp_path, rng):
    """test whether the plot tracker creates files without errors"""
    output_file = tmp_path / "img.png"

    def get_title(state, t):
        return f"{state.integral:g} at {t:g}"

    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid, rng=rng)
    eq = DiffusionPDE()
    tracker = trackers.PlotTracker(
        output_file=output_file,
        title=get_title,
        interval=0.1,
        show=False,
        tight_layout=True,
    )

    eq.solve(state, t_range=0.5, dt=0.005, tracker=tracker, backend="numpy")

    assert output_file.stat().st_size > 0


@pytest.mark.skipif(not Movie.is_available(), reason="no ffmpeg")
def test_plot_movie_tracker(tmp_path, rng):
    """test whether the plot tracker creates files without errors"""
    output_file = tmp_path / "movie.mov"

    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid, rng=rng)
    eq = DiffusionPDE()
    tracker = trackers.PlotTracker(
        movie=output_file, interval=0.1, show=False, tight_layout=True
    )

    eq.solve(state, t_range=0.5, dt=0.005, tracker=tracker, backend="numpy")

    assert output_file.stat().st_size > 0


def test_simple_progress():
    """simple test for basic progress bar"""
    pbar = trackers.ProgressTracker(interval=1)
    field = ScalarField(UnitGrid([3]))
    pbar.initialize(field)
    pbar.handle(field, 2)
    pbar.finalize()


def test_trackers(rng):
    """test whether simple trackers can be used"""
    times = []

    def store_time(state, t):
        times.append(t)

    def get_data(state):
        return {"integral": state.integral}

    devnull = open(os.devnull, "w")
    data = trackers.DataTracker(get_data, interval=0.1)
    tracker_list = [
        trackers.PrintTracker(interval=0.1, stream=devnull),
        trackers.CallbackTracker(store_time, interval=0.1),
        None,  # should be ignored
        data,
    ]
    if module_available("matplotlib"):
        tracker_list.append(trackers.PlotTracker(interval=0.1, show=False))

    grid = UnitGrid([16, 16])
    state = ScalarField.random_uniform(grid, 0.2, 0.3, rng=rng)
    eq = DiffusionPDE()
    eq.solve(state, t_range=1, dt=0.005, tracker=tracker_list)

    devnull.close()

    assert times == data.times
    if module_available("pandas"):
        df = data.dataframe
        np.testing.assert_allclose(df["time"], times)
        np.testing.assert_allclose(df["integral"], state.integral)


def test_callback_tracker(rng):
    """test trackers that support a callback"""
    data = []

    def store_mean_data(state):
        data.append(state.average)

    def get_mean_data(state):
        return state.average

    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid, 0.2, 0.3, rng=rng)
    eq = DiffusionPDE()
    data_tracker = trackers.DataTracker(get_mean_data, interval=0.1)
    callback_tracker = trackers.CallbackTracker(store_mean_data, interval=0.1)
    tracker_list = [data_tracker, callback_tracker]
    eq.solve(state, t_range=0.5, dt=0.005, tracker=tracker_list, backend="numpy")

    np.testing.assert_array_equal(data, data_tracker.data)

    data = []

    def store_time(state, t):
        data.append(t)

    def get_time(state, t):
        return t

    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid, 0.2, 0.3, rng=rng)
    eq = DiffusionPDE()
    data_tracker = trackers.DataTracker(get_time, interval=0.1)
    tracker_list = [trackers.CallbackTracker(store_time, interval=0.1), data_tracker]
    eq.solve(state, t_range=0.5, dt=0.005, tracker=tracker_list, backend="numpy")

    ts = np.arange(0, 0.55, 0.1)
    np.testing.assert_allclose(data, ts, atol=1e-2)
    np.testing.assert_allclose(data_tracker.data, ts, atol=1e-2)


def test_data_tracker(tmp_path):
    """test the DataTracker"""
    field = ScalarField(UnitGrid([4, 4]))
    eq = DiffusionPDE()

    path = tmp_path / "test_data_tracker.pickle"
    data1 = trackers.DataTracker(lambda f: f.average, filename=path)
    data2 = trackers.DataTracker(lambda f: {"avg": f.average, "int": f.integral})
    eq.solve(field, 10, tracker=[data1, data2])

    with path.open("br") as fp:
        time, data = pickle.load(fp)
    np.testing.assert_allclose(time, np.arange(11))
    assert isinstance(data, list)
    assert len(data) == 11

    assert path.stat().st_size > 0


def test_steady_state_tracker():
    """test the SteadyStateTracker"""
    storage = MemoryStorage()
    c0 = ScalarField.from_expression(UnitGrid([5]), "sin(x)")
    eq = DiffusionPDE()

    # use basic form
    tracker = trackers.SteadyStateTracker(atol=0.05, rtol=0.05, progress=True)
    eq.solve(c0, 1e4, dt=0.1, tracker=[tracker, storage.tracker(interval=1e2)])
    assert len(storage) < 20  # finished early

    # use form with the evolution rate supplied
    tracker = trackers.SteadyStateTracker(
        atol=0.05,
        rtol=0.05,
        progress=True,
        evolution_rate=eq.make_pde_rhs(c0, backend="numpy"),
    )
    eq.solve(c0, 1e4, dt=0.1, tracker=[tracker, storage.tracker(interval=1e2)])
    assert len(storage) < 20  # finished early


def test_small_tracker_dt(rng):
    """test the case where the dt of the tracker is smaller than the dt
    of the simulation"""
    storage = MemoryStorage()
    eq = DiffusionPDE()
    c0 = ScalarField.random_uniform(UnitGrid([4, 4]), 0.1, 0.2, rng=rng)
    eq.solve(
        c0, 1e-2, dt=1e-3, solver="explicit", tracker=storage.tracker(interval=1e-4)
    )
    assert len(storage) == 11


def test_runtime_tracker(rng):
    """test the RuntimeTracker"""
    s = ScalarField.random_uniform(UnitGrid([128]), rng=rng)
    tracker = trackers.RuntimeTracker("0:01")
    sol = ExplicitSolver(DiffusionPDE())
    con = Controller(sol, t_range=1e4, tracker=["progress", tracker])
    con.run(s, dt=1e-3)


def test_consistency_tracker(rng):
    """test the ConsistencyTracker"""
    s = ScalarField.random_uniform(UnitGrid([128]), rng=rng)
    sol = ExplicitSolver(DiffusionPDE(1e3))
    con = Controller(sol, t_range=1e5, tracker=["consistency"])
    with np.errstate(all="ignore"):
        con.run(s, dt=1)
    assert con.info["t_final"] < con.info["t_end"]


def test_material_conservation_tracker(rng):
    """test the MaterialConservationTracker"""
    state = ScalarField.random_uniform(UnitGrid([8, 8]), 0, 1, rng=rng)

    solver = ExplicitSolver(CahnHilliardPDE())
    controller = Controller(solver, t_range=1, tracker=["material_conservation"])
    controller.run(state, dt=1e-3)
    assert controller.info["t_final"] >= 1

    solver = ExplicitSolver(AllenCahnPDE())
    controller = Controller(solver, t_range=1, tracker=["material_conservation"])
    controller.run(state, dt=1e-3)
    assert controller.info["t_final"] <= 1


def test_get_named_trackers():
    """test the get_named_trackers function"""
    for name, cls in get_named_trackers().items():
        assert isinstance(name, str)
        tracker = TrackerBase.from_data(name)
        assert isinstance(tracker, cls)


def test_double_tracker(rng):
    """simple test for using a custom tracker twice"""
    interval = ConstantInterrupts(1)
    times1, times2 = [], []
    t1 = trackers.CallbackTracker(lambda s, t: times1.append(t), interval=interval)
    t2 = trackers.CallbackTracker(lambda s, t: times2.append(t), interval=interval)

    field = ScalarField.random_uniform(UnitGrid([3]), rng=rng)
    DiffusionPDE().solve(field, t_range=4, dt=0.1, tracker=[t1, t2])

    np.testing.assert_allclose(times1, np.arange(5))
    np.testing.assert_allclose(times2, np.arange(5))



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import itertools
from copy import copy, deepcopy

import numpy as np
import pytest

from pde import grids
from pde.grids.base import (
    GridBase,
    OperatorInfo,
    discretize_interval,
    registered_operators,
)
from pde.tools.misc import skipUnlessModule


def iter_grids():
    """generator providing some test grids"""
    for periodic in [True, False]:
        yield grids.UnitGrid([3], periodic=periodic)
        yield grids.UnitGrid([3, 3, 3], periodic=periodic)
        yield grids.CartesianGrid([[-1, 2], [0, 3]], [5, 7], periodic=periodic)
        yield grids.CylindricalSymGrid(3, [-1, 2], [7, 8], periodic_z=periodic)
    yield grids.PolarSymGrid(3, 4)
    yield grids.SphericalSymGrid(3, 4)


@pytest.mark.parametrize("grid", iter_grids())
def test_basic_grid_properties(grid):
    """test basic grid properties"""
    with pytest.raises(AttributeError):
        grid.periodic = True
    with pytest.raises(AttributeError):
        grid.shape = 12


def test_discretize(rng):
    """test the discretize function"""
    x_min = rng.uniform(0, 1)
    x_max = rng.uniform(2, 3)
    num = rng.integers(5, 8)
    x, dx = discretize_interval(x_min, x_max, num)
    assert dx == pytest.approx((x_max - x_min) / num)
    x_expect = np.linspace(x_min + dx / 2, x_max - dx / 2, num)
    np.testing.assert_allclose(x, x_expect)


@pytest.mark.parametrize("grid", iter_grids())
def test_serialization(grid):
    """test whether grid can be serialized and copied"""
    g = GridBase.from_state(grid.state_serialized)
    assert grid == g
    assert grid._cache_hash() == g._cache_hash()

    for g in (grid.copy(), copy(grid), deepcopy(grid)):
        assert grid == g
        assert grid is not g


def test_iter_mirror_points():
    """test iterating mirror points in grids"""
    grid_cart = grids.UnitGrid([2, 2], periodic=[True, False])
    grid_cyl = grids.CylindricalSymGrid(2, (0, 2), (2, 2), periodic_z=False)
    grid_sph = grids.SphericalSymGrid(2, 2)
    assert grid_cart._cache_hash() != grid_cyl._cache_hash() != grid_sph._cache_hash()

    for with_, only_periodic in itertools.product([False, True], repeat=2):
        num_expect = 2 if only_periodic else 8
        num_expect += 1 if with_ else 0
        ps = grid_cart.iter_mirror_points([1, 1], with_, only_periodic)
        assert len(list(ps)) == num_expect

        num_expect = 0 if only_periodic else 2
        num_expect += 1 if with_ else 0
        ps = grid_cyl.iter_mirror_points([0, 0, 1], with_, only_periodic)
        assert len(list(ps)) == num_expect

        num_expect = 1 if with_ else 0
        ps = grid_sph.iter_mirror_points([0, 0, 0], with_, only_periodic)
        assert len(list(ps)) == num_expect


@pytest.mark.parametrize("grid", iter_grids())
def test_coordinate_conversion(grid, rng):
    """test the conversion between cells and points"""
    p_empty = np.zeros((0, grid.dim))
    c_empty = np.zeros((0, grid.num_axes))

    p = grid.get_random_point(coords="grid", rng=rng)
    for coords in ["cartesian", "cell", "grid"]:
        # test empty conversion
        assert grid.transform(p_empty, "cartesian", coords).size == 0
        assert grid.transform(c_empty, "grid", coords).size == 0
        assert grid.transform(c_empty, "cell", coords).size == 0

        # test full conversion
        p1 = grid.transform(p, "grid", coords)
        for target in ["cartesian", "grid"]:
            p2 = grid.transform(p1, coords, target)
            p3 = grid.transform(p2, target, coords)
            np.testing.assert_allclose(p1, p3, err_msg=f"{coords} -> {target}")


@pytest.mark.parametrize("grid", iter_grids())
def test_integration_serial(grid, rng):
    """test integration of fields"""
    arr = rng.normal(size=grid.shape)
    res = grid.make_integrator()(arr)
    assert np.isscalar(res)
    assert res == pytest.approx(grid.integrate(arr))
    if grid.num_axes == 1:
        assert res == pytest.approx(grid.integrate(arr, axes=0))
    else:
        assert res == pytest.approx(grid.integrate(arr, axes=range(grid.num_axes)))


@skipUnlessModule("matplotlib")
def test_grid_plotting():
    """test plotting of grids"""
    grids.UnitGrid([4]).plot()
    grids.UnitGrid([4, 4]).plot()

    with pytest.raises(NotImplementedError):
        grids.UnitGrid([4, 4, 4]).plot()

    grids.PolarSymGrid(4, 8).plot()
    grids.PolarSymGrid((2, 4), 8).plot()


@pytest.mark.parametrize("grid", iter_grids())
def test_operators(grid):
    """test operator mechanism"""

    def make_op(state):
        return lambda state: state

    assert "laplace" in grid.operators
    with pytest.raises(ValueError):
        grid.make_operator("not_existent", "auto_periodic_neumann")
    grid.register_operator("noop", make_op)
    assert "noop" in grid.operators
    del grid._operators["noop"]  # reset original state

    # test all instance operators
    for op in grid.operators:
        assert isinstance(grid._get_operator_info(op), OperatorInfo)


def test_cartesian_operator_infos():
    """test special case of cartesian operators"""
    assert "d_dx" not in grids.UnitGrid.operators
    assert "d_dx" in grids.UnitGrid([2]).operators
    assert "d_dy" not in grids.UnitGrid([2]).operators


def test_registered_operators():
    """test the registered_operators function"""
    for grid_name, ops in registered_operators().items():
        grid_class_ops = getattr(grids, grid_name).operators
        assert all(op in grid_class_ops for op in ops)



"""
Spherically symmetric PDE
=========================

This example illustrates how to solve a PDE in a spherically symmetric geometry.
"""

from pde import DiffusionPDE, ScalarField, SphericalSymGrid

grid = SphericalSymGrid(radius=[1, 5], shape=128)  # generate grid
state = ScalarField.random_uniform(grid)  # generate initial condition

eq = DiffusionPDE(0.1)  # define the PDE
result = eq.solve(state, t_range=0.1, dt=0.001)

result.plot(kind="image")



"""
The py-pde package provides classes and methods for solving partial differential
equations.
"""

# determine the package version
try:
    # try reading version of the automatically generated module
    from ._version import __version__
except ImportError:
    # determine version automatically from CVS information
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version("pde")
    except PackageNotFoundError:
        # package is not installed, so we cannot determine any version
        __version__ = "unknown"
    del PackageNotFoundError, version  # clean name space

# initialize the configuration
from .tools.config import Config, environment

config = Config()  # initialize the default configuration
del Config  # clean name space

# import all other modules that should occupy the main name space
from .fields import *  # @UnusedWildImport
from .grids import *  # @UnusedWildImport
from .pdes import *  # @UnusedWildImport
from .solvers import *  # @UnusedWildImport
from .storage import *  # @UnusedWildImport
from .tools.parameters import Parameter
from .trackers import *  # @UnusedWildImport
from .visualization import *  # @UnusedWildImport



"""
Functions and classes for plotting simulation data

.. autosummary::
   :nosignatures:

   ScalarFieldPlot
   plot_magnitudes
   plot_kymograph
   plot_kymographs
   plot_interactive

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import logging
import math
import time
import warnings
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np

from ..fields import FieldCollection
from ..fields.base import DataFieldBase, FieldBase
from ..storage.base import StorageBase
from ..tools.docstrings import fill_in_docstring
from ..tools.misc import module_available
from ..tools.output import display_progress
from ..tools.plotting import (
    PlotReference,
    napari_add_layers,
    napari_viewer,
    plot_on_axes,
    plot_on_figure,
)

_logger = logging.getLogger(__name__)
ScaleData = Union[str, float, Tuple[float, float]]


def _add_horizontal_colorbar(im, ax, num_loc: int = 5) -> None:
    """adds a horizontal colorbar for image `im` to the axis `ax`

    Args:
        im: The result of calling :func:`matplotlib.pyplot.imshow`
        ax: The matplotlib axes to which the colorbar is added
        num_loc (int): Number of ticks
    """
    from matplotlib.ticker import MaxNLocator
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.new_vertical(size="5%", pad=0.1, pack_start=True)
    fig.add_axes(cax)
    cb = fig.colorbar(im, cax=cax, orientation="horizontal")
    cb.locator = MaxNLocator(num_loc)
    cb.update_ticks()


def extract_field(
    fields: FieldBase,
    source: Union[None, int, Callable] = None,
    check_rank: Optional[int] = None,
) -> DataFieldBase:
    """Extracts a single field from a possible collection.

    Args:
        fields (:class:`~pde.fields.FieldBase`):
            The field from which data is extracted
        source (int or callable, optional):
            Determines how a field is extracted from `fields`. If `None`,
            `fields` is passed as is, assuming it is already a scalar field.
            This works for the simple, standard case where only a single
            :class:`~pde.fields.scalar.ScalarField` is treated. Alternatively,
            `source` can be an integer, indicating which field is extracted from
            an instance of :class:`~pde.fields.FieldCollection`.
            Lastly, `source` can be a function that takes `fields` as an
            argument and returns the desired field.
        check_rank (int, optional):
            Can be given to check whether the extracted field has the correct
            rank (0 = ScalarField, 1 = VectorField, ...).

    Returns:
        :class:`~pde.fields.DataFieldBase`: The extracted field
    """
    # extract the source
    if source is None:
        field = fields
    elif callable(source):
        field = source(fields)
    else:
        if isinstance(fields, FieldCollection):
            field = fields[source]
        else:
            raise TypeError(
                f"Cannot extract component {source} from instance of "
                f"{fields.__class__.__name__}"
            )

    if isinstance(field, FieldCollection):
        raise RuntimeError("Extract field is a collection")

    # test whether the rank is as expected
    if check_rank is not None and field.rank != check_rank:  # type: ignore
        raise RuntimeError(f"Rank of extracted field is not {check_rank}")

    return field  # type: ignore


class ScalarFieldPlot:
    """class managing compound plots of scalar fields"""

    @fill_in_docstring
    def __init__(
        self,
        fields: FieldBase,
        quantities=None,
        scale: ScaleData = "automatic",
        fig=None,
        title: Optional[str] = None,
        tight: bool = False,
        show: bool = True,
    ):
        """
        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                Collection of fields
            quantities:
                {ARG_PLOT_QUANTITIES}
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
            fig (:class:`matplotlib.figure.Figure):
                Figure to be used for plotting. If `None`, a new figure is
                created.
            title (str):
                Title of the plot.
            tight (bool):
                Whether to call :func:`matplotlib.pyplot.tight_layout`. This
                affects the layout of all plot elements.
            show (bool):
                Flag determining whether to show a plot. If `False`, the plot is
                kept in the background, which can be useful if it only needs to
                be written to a file.
        """
        self.grid = fields.grid
        self.quantities = self._prepare_quantities(fields, quantities, scale=scale)
        self.show = show

        # figure out whether plots are shown in jupyter notebook
        try:
            from ipywidgets import Output
        except ImportError:
            ipython_plot = False
        else:
            ipython_plot = self.show

        if ipython_plot:
            # plotting is done in an ipython environment using widgets
            from IPython.display import display

            self._ipython_out = Output()
            with self._ipython_out:
                self._initialize(fields, scale, fig, title, tight)
            display(self._ipython_out)

        else:
            # plotting is done using a simple matplotlib backend
            self._ipython_out = None
            self._initialize(fields, scale, fig, title, tight)

        if self.show:
            self._show()

    @classmethod
    @fill_in_docstring
    def from_storage(
        cls,
        storage: StorageBase,
        quantities=None,
        scale: ScaleData = "automatic",
        tight: bool = False,
        show: bool = True,
    ) -> "ScalarFieldPlot":
        """create ScalarFieldPlot from storage

        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                Instance of the storage class that contains the data
            quantities:
                {ARG_PLOT_QUANTITIES}
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
            tight (bool):
                Whether to call :func:`matplotlib.pyplot.tight_layout`. This
                affects the layout of all plot elements.
            show (bool):
                Flag determining whether to show a plot. If `False`, the plot is
                kept in the background.

        Returns:
            :class:`~pde.visualization.plotting.ScalarFieldPlot`
        """
        fields = storage[0]
        assert isinstance(fields, FieldBase), "Storage must contain fields"

        # prepare the data that needs to be plotted
        quantities = cls._prepare_quantities(fields, quantities=quantities, scale=scale)

        # resolve automatic scaling
        for quantity_row in quantities:
            for quantity in quantity_row:
                if quantity.get("scale", "automatic") == "automatic":
                    vmin, vmax = +math.inf, -math.inf
                    for data in storage:
                        field = extract_field(data, quantity.get("source"))
                        img = field.get_image_data()
                        vmin = np.nanmin([np.nanmin(img["data"]), vmin])
                        vmax = np.nanmax([np.nanmax(img["data"]), vmax])
                    quantity["scale"] = (vmin, vmax)

        # actually setup
        return cls(fields, quantities, tight=tight, show=show)

    @staticmethod
    @fill_in_docstring
    def _prepare_quantities(
        fields: FieldBase, quantities, scale: ScaleData = "automatic"
    ) -> List[List[Dict[str, Any]]]:
        """internal method to prepare quantities

        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                The field containing the data to show
            quantities (dict):
                {ARG_PLOT_QUANTITIES}
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}

        Returns:
            list of list of dict: a 2d arrangements of panels that define what
                quantities are shown.
        """
        if quantities is None:
            # show each component by default
            if isinstance(fields, FieldCollection):
                quantities = []
                for i, field in enumerate(fields):
                    title = field.label if field.label else f"Field {i + 1}"
                    quantity = {"title": title, "source": i, "scale": scale}
                    quantities.append(quantity)
            else:
                quantities = [{"title": "Concentration", "source": None}]

        # make sure panels is a 2d array
        if isinstance(quantities, dict):
            quantities = [[quantities]]
        elif isinstance(quantities[0], dict):
            quantities = [quantities]

        # set the scaling for quantities where it is not yet set
        for row in quantities:
            for col in row:
                col.setdefault("scale", scale)

        return quantities  # type: ignore

    def __del__(self):
        try:
            if hasattr(self, "fig") and self.fig:
                import matplotlib.pyplot as plt

                plt.close(self.fig)
        except Exception:
            pass  # can occur when shutting down python interpreter

    @fill_in_docstring
    def _initialize(
        self,
        fields: FieldBase,
        scale: ScaleData = "automatic",
        fig=None,
        title: Optional[str] = None,
        tight: bool = False,
    ):
        """initialize the plot creating the figure and the axes

        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                Collection of fields
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
            fig (:class:`matplotlib.figure.Figure):
                Figure to be used for plotting. If `None`, a new figure is
                created.
            title (str):
                Title of the plot.
            tight (bool):
                Whether to call :func:`matplotlib.pyplot.tight_layout`. This
                affects the layout of all plot elements.
        """
        import matplotlib.cm as cm
        import matplotlib.pyplot as plt

        num_rows = len(self.quantities)
        num_cols = max(len(p) for p in self.quantities)

        example_image = fields.get_image_data()

        # set up the figure
        if fig is None:
            figsize = (4 * num_cols, 4 * num_rows)
            self.fig, self.axes = plt.subplots(
                num_rows, num_cols, sharey=True, squeeze=False, figsize=figsize
            )
        else:
            self.fig = fig
            self.axes = fig.axes

        if title is None:
            title = ""
        self.sup_title = plt.suptitle(title)

        # set up all images
        empty = np.zeros_like(example_image["data"])

        self.images: List[List[Any]] = []
        for i, panel_row in enumerate(self.quantities):
            img_row = []
            for j, panel in enumerate(panel_row):
                # determine scale of the panel
                panel_scale = panel.get("scale", scale)
                if panel_scale == "automatic":
                    vmin, vmax = None, None
                elif panel_scale == "unity":
                    vmin, vmax = 0, 1
                elif panel_scale == "symmetric":
                    vmin, vmax = -1, 1
                else:
                    try:
                        vmin, vmax = panel_scale
                    except TypeError:
                        vmin, vmax = 0, panel_scale

                # determine colormap
                cmap = panel.get("cmap")
                if cmap is None:
                    if vmin is None or vmax is None:
                        cmap = cm.viridis
                    elif np.isclose(-vmin, vmax):
                        cmap = cm.coolwarm
                    elif np.isclose(vmin, 0):
                        cmap = cm.gray
                    else:
                        cmap = cm.viridis

                # initialize image
                ax = self.axes[i, j]
                img = ax.imshow(
                    empty,
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                    extent=example_image["extent"],
                    origin="lower",
                    interpolation="quadric",
                )
                _add_horizontal_colorbar(img, ax)
                ax.set_title(panel.get("title", ""))
                ax.set_xlabel(example_image["label_x"])
                ax.set_ylabel(example_image["label_y"])

                # store the default value alongside
                img._default_vmin_vmax = vmin, vmax
                img_row.append(img)
            self.images.append(img_row)

        if tight:
            # adjust layout and leave some room for title
            self.fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    def _update_data(self, fields: FieldBase, title: Optional[str] = None) -> None:
        """update the fields in the current plot

        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                The field or field collection of which the defined quantities
                are shown.
            title (str, optional):
                The title of this view. If `None`, the current title is not
                changed.
        """
        assert isinstance(fields, FieldBase), "Fields must inherit from FieldBase"

        if title:
            self.sup_title.set_text(title)

        # iterate over all panels and update their content
        for i, panel_row in enumerate(self.quantities):
            for j, panel in enumerate(panel_row):
                # obtain image data
                field = extract_field(fields, panel.get("source"))
                img_data = field.get_image_data()

                # set the data in the correct panel
                img = self.images[i][j]
                img.set_data(img_data["data"])
                vmin, vmax = img._default_vmin_vmax
                if vmin is None:
                    vmin = img_data["data"].min()
                if vmax is None:
                    vmax = img_data["data"].max()
                img.set_clim(vmin, vmax)

    def _show(self):
        """show the updated plot"""
        if self._ipython_out:
            # seems to be in an ipython instance => update widget
            from IPython.display import clear_output, display

            with self._ipython_out:
                clear_output(wait=True)
                display(self.fig)

            # add a small pause to allow the GUI to run it's event loop
            time.sleep(0.01)

        else:
            # seems to be in a normal matplotlib window => update it
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # add a small pause to allow the GUI to run it's event loop
                import matplotlib.pyplot as plt

                plt.pause(0.01)

    def update(self, fields: FieldBase, title: Optional[str] = None) -> None:
        """update the plot with the given fields

        Args:
            fields:
                The field or field collection of which the defined quantities
                are shown.
            title (str, optional):
                The title of this view. If `None`, the current title is not
                changed.
        """
        self._update_data(fields, title)
        if self.show:
            self._show()

    def savefig(self, path: str, **kwargs):
        """save plot to file

        Args:
            path (str):
                The path to the file where the image is written. The file
                extension determines the image format
            **kwargs:
                Additional arguments are forwarded to
                :meth:`matplotlib.figure.Figure.savefig`.
        """
        self.fig.savefig(path, **kwargs)

    def make_movie(
        self, storage: StorageBase, filename: str, progress: bool = True
    ) -> None:
        """make a movie from the data stored in storage

        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                The storage instance that contains all the data for the movie
            filename (str):
                The filename to which the movie is written. The extension
                determines the format used.
            progress (bool):
                Flag determining whether the progress of making the movie is
                shown.
        """
        from ..visualization.movies import Movie

        # create the iterator over the data
        if progress:
            data_iter = display_progress(storage.items(), total=len(storage))
        else:
            data_iter = storage.items()

        with Movie(filename=filename) as movie:
            # iterate over all time steps
            for t, data in data_iter:
                self.update(data, title=f"Time {t:g}")
                movie.add_figure(self.fig)


@plot_on_axes()
@fill_in_docstring
def plot_magnitudes(
    storage: StorageBase, quantities=None, ax=None, **kwargs
) -> PlotReference:
    r"""plot spatially averaged quantities as a function of time

    For scalar fields, the default is to plot the average value while the averaged norm
    is plotted for vector fields.

    Args:
        storage:
            Instance of :class:`~pde.storage.base.StorageBase` that contains
            the simulation data that will be plotted
        quantities:
            {ARG_PLOT_QUANTITIES}
        {PLOT_ARGS}
        \**kwargs:
            All remaining parameters are forwarded to the `ax.plot` method

    Returns:
        :class:`~pde.tools.plotting.PlotReference`: The reference to the plot
    """
    if quantities is None:
        fields = storage[0]
        assert isinstance(fields, FieldBase), "Storage must contain fields"
        if fields.label:
            label_base = fields.label
        else:
            label_base = kwargs.get("label", "Field")

        if isinstance(fields, FieldCollection):
            quantities = []
            for i, field in enumerate(fields):
                if field.label:
                    label = field.label
                else:
                    label = label_base + f" {i + 1}"
                quantities.append({"label": label, "source": i})
        else:
            quantities = [{"label": label_base, "source": None}]

    _logger.debug("Quantities: %s", quantities)

    # prepare data field
    data = [
        {"label": quantity.get("label", ""), "values": []} for quantity in quantities
    ]

    # calculate the data
    for fields in storage:
        for i, quantity in enumerate(quantities):
            source = quantity["source"]
            if callable(source):
                value = source(fields)
            elif source is None:
                value = fields.magnitude  # type: ignore
            else:
                value = fields[source].magnitude  # type: ignore
            value *= quantity.get("scale", 1)
            data[i]["values"].append(value)

    # plot the data
    lines = []
    times = np.asarray(storage.times)
    check_complex = True
    for d in data:
        kwargs["label"] = d["label"]

        # warn if there is an imaginary part
        if check_complex and np.any(np.iscomplex(d["values"])):
            _logger.warning("Only the real part of the complex data is shown")
            check_complex = False  # only warn once

        # actually plot the data
        (l,) = ax.plot(times, np.real(d["values"]), **kwargs)
        lines.append(l)

    ax.set_xlabel("Time")
    if len(data) == 1:
        ax.set_ylabel(kwargs["label"])

    if len(data) > 1:
        ax.legend(loc="best")

    return PlotReference(ax, lines)


def _plot_kymograph(
    img_data: Dict[str, Any],
    ax,
    colorbar: bool = True,
    transpose: bool = False,
    **kwargs,
) -> PlotReference:
    r"""plots a simple kymograph from given data

    Args:
        img_data (dict):
            Contains the kymograph data
        ax (:class:`~matplotlib.axes.Axes`):
            The axes to which the plot is added
        colorbar (bool):
            Whether to show a colorbar or not
        transpose (bool):
            Determines whether the transpose of the data should is plotted
        \**kwargs:
            Additional keyword arguments are passed to
            :func:`matplotlib.pyplot.imshow`.

    Returns:
        :class:`~pde.tools.plotting.PlotReference`: The reference to the plot
    """
    # transpose data if requested
    if transpose:
        # avoid changing the img_data that was passed into the function
        extent = np.r_[img_data["extent_y"], img_data["extent_x"]]
        img_data = {
            "data": img_data["data"].T,
            "label_x": img_data["label_y"],
            "label_y": img_data["label_x"],
        }
    else:
        extent = np.r_[img_data["extent_x"], img_data["extent_y"]]

    # warn if there is an imaginary part
    if np.any(np.iscomplex(img_data["data"])):
        _logger.warning("Only the real part of the complex data is shown")

    # create the actual plot
    axes_image = ax.imshow(
        img_data["data"].real, extent=extent, origin="lower", **kwargs
    )

    # adjust some settings
    ax.set_xlabel(img_data["label_x"])
    ax.set_ylabel(img_data["label_y"])
    ax.set_aspect("auto")

    if colorbar:
        import matplotlib.pyplot as plt

        plt.colorbar(axes_image, ax=ax)
        # Note that we here do not use the method `add_scaled_colorbar` from the
        # `pde.tools.plotting` module since this does not play well with axes with an
        # explicit aspect ratio.

    return PlotReference(ax, axes_image)


@plot_on_axes()
def plot_kymograph(
    storage: StorageBase,
    field_index: Optional[int] = None,
    scalar: str = "auto",
    extract: str = "auto",
    colorbar: bool = True,
    transpose: bool = False,
    ax=None,
    **kwargs,
) -> PlotReference:
    r"""plots a single kymograph from stored data

    The kymograph shows line data stacked along time. Consequently, the
    resulting image shows space along the horizontal axis and time along the
    vertical axis.

    Args:
        storage (:class:`~droplets.simulation.storage.StorageBase`):
            The storage instance that contains all the data
        field_index (int):
            An index to choose a single field out of many in a collection
            stored in `storage`. This option should not be used if only a single
            field is stored in a collection.
        scalar (str):
            The method for extracting scalars as described in
            :meth:`DataFieldBase.to_scalar`.
        extract (str):
            The method used for extracting the line data. See the docstring
            of the grid method `get_line_data` to find supported values.
        colorbar (bool):
            Whether to show a colorbar or not
        transpose (bool):
            Determines whether the transpose of the data should is plotted
        {PLOT_ARGS}
        \**kwargs:
            Additional keyword arguments are passed to
            :func:`matplotlib.pyplot.imshow`.

    Returns:
        :class:`~pde.tools.plotting.PlotReference`: The reference to the plot
    """
    if len(storage) == 0:
        raise RuntimeError("Storage is empty")
    line_data_args: Dict[str, Any] = {"scalar": scalar, "extract": extract}

    if storage.has_collection:
        # there are many fields in the storage
        if field_index is None:
            _logger.warning(
                "Showing first field since argument `field_index` is missing. Use "
                "`plot_kymographs` to show kymographs of multiple fields."
            )
        else:
            line_data_args["index"] = field_index

    elif field_index is not None and field_index != 0:
        # there is only one field in the storage, but a field_index > 0 is given
        _logger.warning("`field_index` should only be set for FieldCollections")

    # prepare the image data for one kymograph
    image = []
    for _, field in storage.items():
        img_data = field.get_line_data(**line_data_args)
        image.append(img_data["data_y"])

    img_data["data"] = np.array(image)
    img_data["extent_y"] = (storage.times[0], storage.times[-1])
    img_data["label_y"] = "Time"

    return _plot_kymograph(
        img_data, ax, colorbar=colorbar, transpose=transpose, **kwargs
    )


@plot_on_figure()
def plot_kymographs(
    storage: StorageBase,
    scalar: str = "auto",
    extract: str = "auto",
    colorbar: bool = True,
    transpose: bool = False,
    resize_fig: bool = True,
    fig=None,
    **kwargs,
) -> List[PlotReference]:
    r"""plots kymographs for all fields stored in `storage`

    The kymograph shows line data stacked along time. Consequently, the
    resulting image shows space along the horizontal axis and time along the
    vertical axis.

    Args:
        storage (:class:`~droplets.simulation.storage.StorageBase`):
            The storage instance that contains all the data
        scalar (str):
            The method for extracting scalars as described in
            :meth:`DataFieldBase.to_scalar`.
        extract (str):
            The method used for extracting the line data. See the docstring
            of the grid method `get_line_data` to find supported values.
        colorbar (bool):
            Whether to show a colorbar or not
        transpose (bool):
            Determines whether the transpose of the data should is plotted
        resize_fig (bool):
            Whether to resize the figure to adjust to the number of panels
        {PLOT_ARGS}
        \**kwargs:
            Additional keyword arguments are passed to the calls to
            :func:`matplotlib.pyplot.imshow`.

    Returns:
        list of :class:`~pde.tools.plotting.PlotReference`: The references to
        all plots
    """
    if len(storage) == 0:
        raise RuntimeError("Storage is empty")
    line_data_args = {"scalar": scalar, "extract": extract}
    if storage.has_collection:
        num_fields = len(storage[0])  # type: ignore
    else:
        num_fields = 1

    # prepare the image data for one kymograph
    images = []
    for _, field in storage.items():
        if storage.has_collection:
            image_lines = []
            for f_id in range(num_fields):
                img_data = field.get_line_data(  # type: ignore
                    index=f_id,
                    **line_data_args,
                )
                image_lines.append(img_data["data_y"])
        else:
            img_data = field.get_line_data(**line_data_args)
            image_lines = [img_data["data_y"]]
        images.append(image_lines)

    img_arr = np.array(images)
    img_data["extent_y"] = (storage.times[0], storage.times[-1])
    img_data["label_y"] = "Time"

    # disable interactive plotting temporarily
    # create a plot with all the panels
    if resize_fig:
        fig.set_size_inches((4 * num_fields, 3), forward=True)
    axs = fig.subplots(1, num_fields, squeeze=False)

    # iterate over all axes and plot the kymograph
    refs = []
    for i, ax in enumerate(axs[0]):
        img_data["data"] = img_arr[:, i]
        ref = _plot_kymograph(
            img_data, ax, colorbar=colorbar, transpose=transpose, **kwargs
        )
        if storage.has_collection:
            ax.set_title(storage._field[i].label)  # type: ignore
        refs.append(ref)

    return refs


def plot_interactive(
    storage: StorageBase,
    time_scaling: Literal["exact", "scaled"] = "exact",
    viewer_args: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    r"""plots stored data interactively using the `napari <https://napari.org>`_ viewer

    Args:
        storage (:class:`~droplets.simulation.storage.StorageBase`):
            The storage instance that contains all the data
        time_scaling (str):
            Defines how the time axis is scaled. Possible options are "exact" (the
            actual time points are used), or "scaled" (the axis is scaled so that it has
            similar dimension to the spatial axes). Note that the spatial axes will
            never be scaled.
        viewer_args (dict):
            Arguments passed to :class:`napari.viewer.Viewer` to affect the viewer.
        **kwargs:
            Extra arguments passed to the plotting function
    """
    if not module_available("napari"):
        raise ImportError("Require the `napari` module for interactive plotting")

    if len(storage) < 1:
        raise ValueError("Storage is empty")

    grid = storage.grid
    if grid is None:
        raise RuntimeError("Storage did not contain information about the grid")

    # collect data from all time points
    timecourse: Dict[str, List[np.ndarray]] = dict()
    for field in storage:
        layer_data = field._get_napari_data(**kwargs)

        if timecourse:
            for key, field_data in layer_data.items():
                timecourse[key].append(field_data["data"])
        else:
            for key, field_data in layer_data.items():
                timecourse[key] = [field_data["data"]]

    # replace the data in the layer_data
    for key, field_data in layer_data.items():
        field_data["data"] = np.array(timecourse[key])
        if "scale" in field_data:
            if time_scaling == "exact":
                dt = np.diff(storage.times)[0] if len(storage) > 1 else 1
            elif time_scaling == "scaled":
                length_scale = grid.volume ** (1 / grid.dim)
                dt = length_scale / len(storage)
            else:
                raise ValueError(f"Unknown time scaling `{time_scaling}`")
            field_data["scale"] = np.r_[dt, field_data["scale"]]

    if viewer_args is None:
        viewer_args = {}
    viewer_args.setdefault("axis_labels", ["Time"] + grid.axes)

    # actually display the data using napari
    with napari_viewer(grid, **viewer_args) as viewer:
        napari_add_layers(viewer, layer_data)



"""
Using an interactive tracker
============================

This example illustrates how a simulation can be analyzed live using the
`napari <https://napari.org>`_ viewer.
"""

import pde


def main():
    grid = pde.UnitGrid([64, 64])
    field = pde.ScalarField.random_uniform(grid, label="Density")

    eq = pde.CahnHilliardPDE()
    eq.solve(field, t_range=1e3, dt=1e-3, tracker=["progress", "interactive"])


if __name__ == "__main__":
    # this safeguard is required since the interactive tracker uses multiprocessing
    main()



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import MemoryStorage, UnitGrid
from pde.fields import FieldCollection, ScalarField, Tensor2Field, VectorField


def test_memory_storage():
    """test methods specific to memory storage"""
    sf = ScalarField(UnitGrid([1]))
    s1 = MemoryStorage()
    s1.start_writing(sf)
    sf.data = 0
    s1.append(sf, 0)
    sf.data = 2
    s1.append(sf, 1)

    s2 = MemoryStorage()
    s2.start_writing(sf)
    sf.data = 1
    s2.append(sf, 0)
    sf.data = 3
    s2.append(sf, 1)

    # test from_fields
    s3 = MemoryStorage.from_fields(s1.times, [s1[0], s1[1]])
    assert s3.times == s1.times
    np.testing.assert_allclose(s3.data, s1.data)

    # test from_collection
    s3 = MemoryStorage.from_collection([s1, s2])
    assert s3.times == s1.times
    np.testing.assert_allclose(np.ravel(s3.data), np.arange(4))


def test_field_type_guessing(rng):
    """test the ability to guess the field type"""
    for cls in [ScalarField, VectorField, Tensor2Field]:
        grid = UnitGrid([3])
        field = cls.random_normal(grid, rng=rng)
        s = MemoryStorage()
        s.start_writing(field)
        s.append(field, 0)
        s.append(field, 1)

        # delete information
        s._field = None
        s.info = {}

        assert not s.has_collection
        assert len(s) == 2
        assert s[0] == field

    field = FieldCollection([ScalarField(grid), VectorField(grid)])
    s = MemoryStorage()
    s.start_writing(field)
    s.append(field, 0)

    assert s.has_collection

    # delete information
    s._field = None
    s.info = {}

    with pytest.raises(RuntimeError):
        s[0]
